# coding=utf-8
"""
Reusable widgets to be included in views.

NOTE: code is currently quite messy. Needs to be refactored.
"""
from __future__ import absolute_import, division, print_function

import base64
import cgi
import logging
import re
import urlparse
from collections import namedtuple
from datetime import datetime
from itertools import ifilter

import bleach
import sqlalchemy as sa
import wtforms
from flask import (Markup, current_app, g, json, render_template,
                   render_template_string)
from flask_babel import format_date, format_datetime, format_number, get_locale
from flask_login import current_user
from flask_wtf.file import FileField
from future.utils import string_types
from wtforms.widgets import PasswordInput as BasePasswordInput
from wtforms.widgets import TextArea as BaseTextArea
from wtforms.widgets import HTMLString, Input, Select, html_params
from wtforms_alchemy import ModelFieldList

from abilian.core.entities import Entity
from abilian.core.models.blob import Blob
from abilian.i18n import _, _l
from abilian.services import image
from abilian.services.image import get_format
from abilian.web import csrf, url_for
from abilian.web.filters import babel2datepicker, labelize

from .util import babel2datetime

logger = logging.getLogger(__name__)

__all__ = ['linkify_url', 'text2html', 'Column', 'BaseTableView',
           'MainTableView', 'RelatedTableView', 'AjaxMainTableView',
           'SingleView', 'Panel', 'Row', 'Chosen', 'TagInput',
           'DateInput', 'DefaultViewWidget', 'BooleanWidget', 'FloatWidget',
           'DateTimeWidget', 'DateWidget', 'MoneyWidget', 'EmailWidget',
           'URLWidget', 'ListWidget', 'TabularFieldListWidget',
           'ModelListWidget', 'Select2', 'Select2Ajax', 'RichTextWidget',
           'FileInput', 'EntityWidget']


def linkify_url(value):
  """Tranform an URL pulled from the database to a safe HTML fragment."""

  value = value.strip()

  rjs = r'[\s]*(&#x.{1,7})?'.join(list('javascript:'))
  rvb = r'[\s]*(&#x.{1,7})?'.join(list('vbscript:'))
  re_scripts = re.compile('(%s)|(%s)' % (rjs, rvb), re.IGNORECASE)

  value = re_scripts.sub('', value)

  url = value
  if not url.startswith("http://") and not url.startswith("https://"):
    url = "http://" + url

  url = urlparse.urlsplit(url).geturl()
  if '"' in url:
    url = url.split('"')[0]
  if '<' in url:
    url = url.split('<')[0]

  if value.startswith("http://"):
    value = value[len("http://"):]
  elif value.startswith("https://"):
    value = value[len("https://"):]

  if value.count("/") == 1 and value.endswith("/"):
    value = value[0:-1]

  return '<a href="%s">%s</a>&nbsp;<i class="fa fa-external-link"></i>' % (url, value)


def text2html(text):
  text = text.strip()
  if re.search('<(p|br)>', text.lower()):
    return text
  if '\n' not in text:
    return text

  lines = text.split("\n")
  lines = [ line for line in lines if line ]
  paragraphs = ['<p>%s</p>' % line for line in lines]
  return Markup(bleach.clean("\n".join(paragraphs), tags=['p']))


class Column(object):

  def __init__(self, **kw):
    for k, w in kw.items():
      setattr(self, k, w)


# TODO: rewrite
class BaseTableView(object):
  """
  """
  show_controls = False
  show_search = None
  paginate = False
  options = {}

  def __init__(self, columns, options=None):
    if self.show_search is None:
      self.show_search = self.show_controls

    self.init_columns(columns)
    self.name = u'{}-{:d}'.format(self.__class__.__name__.lower(),
                                  next(g.id_generator))
    if options is not None:
      self.options = options
      self.show_controls = self.options.get('show_controls', self.show_controls)
      self.show_search = self.options.get('show_search', self.show_controls)
      self.paginate = self.options.get('paginate', self.paginate)

  def init_columns(self, columns):
    # TODO
    self.columns = []
    default_width = '{:2.0f}%'.format(0.99 / len(columns) * 100)
    for col in columns:
      if isinstance(col, string_types):
        col = dict(name=col, width=default_width)
      assert type(col) == dict
      col.setdefault('width', default_width)
      col.setdefault('sorting', ('asc', 'desc'))
      if 'label' not in col:
        col['label'] = labelize(col['name'])
      self.columns.append(col)

  def render(self, entities, **kwargs):
    aoColumns = []
    aaSorting = []
    offset = 0
    if self.show_controls:
      aoColumns.append({'asSorting': [] })
      offset = 1

    for idx, c in enumerate(self.columns, offset):
      aoColumns.append({'asSorting': c['sorting'],
                       'sWidth': str(c['width'])})
      aaSorting.append([idx, c['sorting'][0]])

    datatable_options = {
      'aaSorting': aaSorting,
      'aoColumns': aoColumns,
      'bFilter': self.show_search,
      'oLanguage': {
        'sSearch': self.options.get('search_label', _("Filter records:")),
      },
      'bStateSave': False,
      'bPaginate': self.paginate,
      'sPaginationType': "bootstrap",
      'bLengthChange': False,
      'iDisplayLength': self.options.get('paginate_length', 50)
    }
    js = render_template_string(
        '''
        require(
            ['jquery', 'jquery.dataTables'],
            function($) {
                $('#{{ table_id  }}').dataTable({{ options|tojson|safe }});
            });
        ''',
        table_id=self.name,
        options=datatable_options,
    )

    table = []
    for entity in entities:
      table.append(self.render_line(entity))

    template = filter(bool, (self.options.get('template'),
                             'widgets/render_table.html'))
    return Markup(render_template(template,
                                  table=table, js=Markup(js), view=self,
                                  **kwargs))

  def render_line(self, entity):
    line = []
    make_link_on = self.options.get("make_link_on")

    for col in self.columns:
      if type(col) == str:
        column_name = col
        build_url = url_for
      else:
        column_name = col['name']
        build_url = col.get('url', url_for)

      value = entity
      attr = column_name.rsplit('.', 1)
      if len(attr) > 1:
        try:
          value = getattr(value, attr[0], "")
          value = value.display_value(attr[1])
        except AttributeError:
          value = ""
      else:
        value = value.display_value(attr[0])

      format = col.get('format')
      if format:
        cell = format(value)
      elif column_name in (make_link_on, 'name') \
          or col.get('linkable'):
        cell = Markup('<a href="%s">%s</a>'
                      % (build_url(entity), cgi.escape(unicode(value))))
      elif isinstance(value, Entity):
        cell = Markup('<a href="%s">%s</a>'
                      % (build_url(value), cgi.escape(value.name)))
      elif isinstance(value, string_types) \
          and (value.startswith("http://") or value.startswith("www.")):
        cell = Markup(linkify_url(value))
      elif value in (True, False):
        cell = u'\u2713' if value else u''  # Unicode "Check mark"
      elif isinstance(value, list):
        cell = "; ".join(value)
      else:
        if not isinstance(value, (Markup,) + string_types):
          if value is None:
            value = u''
          else:
            value = unicode(value)
        cell = value

      line.append(cell)
    return line


class MainTableView(BaseTableView):
  """
  Table view for main objects list.
  """
  show_controls = True
  paginate = True


class RelatedTableView(BaseTableView):
  """
  Table view for related objects list.
  """
  show_controls = False
  paginate = False


class AjaxMainTableView(object):
  """
  Variant of the MainTableView that gets content from AJAX requests.

  TODO: refactor all of this (currently code is copy/pasted!).
  """
  show_controls = False
  paginate = True
  options = {}

  def __init__(self, columns, ajax_source, search_criterions=(), name=None, options=None):
    self.init_columns(columns)
    self.ajax_source = ajax_source
    self.search_criterions = search_criterions
    self.name = name if name is not None else id(self)
    self.save_state = name is not None
    if options is not None:
      self.options = options

  def init_columns(self, columns):
    # TODO: compute the correct width for each column.
    self.columns = []
    default_width = 0.99 / len(columns)
    for col in columns:
      if type(col) == str:
        col = dict(name=col, width=default_width)
      assert type(col) == dict
      if 'label' not in col:
        col['label'] = labelize(col['name'])

      col.setdefault('sorting', ["asc", "desc"])

      if not col['sorting']:
        col.setdefault('sortable', False)
      else:
        col.setdefault('sortable', True)

      self.columns.append(col)

  def render(self):
    aoColumns = [{'asSorting': []}] if self.show_controls else []
    aoColumns += [{'asSorting': col['sorting'], 'bSortable': col['sortable']}
                  for col in self.columns]
    datatable_options = {
      'sDom': 'fFriltip',
      'aoColumns': aoColumns,
      'bFilter': True,
      'oLanguage': {
        'sSearch': self.options.get('search_label', _(u'Filter records:')),
        'oPaginate': {
          'sPrevious': _(u'Previous'),
          'sNext': _(u'Next'),
        },
        'sLengthMenu': _(u'Entries per page: _MENU_'),
        'sInfo': _(u'Showing _START_ to _END_ of _TOTAL_ entries'),
        'sInfoEmpty': _(u'Showing _START_ to _END_ of _TOTAL_ entries'),
        'sInfoFiltered': _(u'(filtered from _MAX_ total entries)'),
        'sAddAdvancedFilter': _(u'Add a filter'),
        'sZeroRecords': _(u'No matching records found'),
        'sEmptyTable': _(u'No matching records found'),
      },
      'bPaginate': self.paginate,
      'sPaginationType': "bootstrap",
      'bLengthChange': True,
      'iDisplayLength': 25,
      'bStateSave': self.save_state,
      'bProcessing': True,
      'bServerSide': True,
      'sAjaxSource': self.ajax_source,
    }
    if self.options.get('aaSorting', None):
      datatable_options['aaSorting'] = self.options.get('aaSorting')

    advanced_search_filters = []
    for c in self.search_criterions:
      if not c.has_form_filter:
        continue
      d = dict(name=c.name,
               label=unicode(c.label),
               type=c.form_filter_type,
               args=c.form_filter_args,
               unset=c.form_unset_value,)
      if c.has_form_default_value:
        d['defaultValue'] = c.form_default_value

      advanced_search_filters.append(d)

    if advanced_search_filters:
      datatable_options['aoAdvancedSearchFilters'] = advanced_search_filters

    return Markup(render_template('widgets/render_ajax_table.html',
                                  datatable_options=datatable_options,
                                  view=self))

  def render_line(self, entity):
    line = []
    for col in self.columns:
      if type(col) == str:
        column_name = col
      else:
        column_name = col['name']

      value = entity.display_value(column_name)
      cell = None
      has_custom_display = False
      # Manual massage.
      # 'display_fun' gets value *and* entity: useful to perform
      # specific markup based on other entity values.
      # 'display_fmt' is for simple value formatting (format_date from
      # babel for example)
      if value is None:
        value = ""
      elif 'display_fun' in col:
        has_custom_display = True
        value = col['display_fun'](entity, value)
      elif 'display_fmt' in col:
        value = col['display_fmt'](value)

      if has_custom_display:
        cell = value
      elif column_name == 'name':
        cell = Markup('<a href="%s">%s</a>'
                      % (url_for(entity), cgi.escape(value)))
      elif isinstance(value, Entity):
        cell = Markup('<a href="%s">%s</a>'
                      % (url_for(value), cgi.escape(value.name)))
      elif (isinstance(value, string_types)
            and (value.startswith("http://") or value.startswith("www."))):
        cell = Markup(linkify_url(value))
      elif col.get('linkable'):
        cell = Markup('<a href="%s">%s</a>'
                      % (url_for(entity), cgi.escape(unicode(value))))
      else:
        cell = unicode(value)

      line.append(cell)
    return line


#
# Single object view
#
class SingleView(object):
  """View on a single object."""

  def __init__(self, form, *panels, **options):
    self.form = form
    self.panels = panels
    self.options = options

  def render(self, item, form, related_views=()):
    mapper = sa.orm.class_mapper(item.__class__)
    panels = []
    _to_skip = (None, False, 0, 0.0, '', u'-')

    for panel in self.panels:
      data = {}
      field_name_iter = (fn for row in panel.rows
                         for fn in row)

      for name in field_name_iter:
        field = form._fields[name]
        if field.is_hidden:
          continue

        value = field.data
        if not isinstance(field, FileField) and not field.flags.render_empty:
          if value in _to_skip:
            continue

        value = Markup(field.render_view(entity=item))
        if value == u'':
          # related models may have [] as value, but we don't discard this type
          # of value in order to let widget a chance to render something useful
          # like an 'add model' button.
          #
          # if it renders an empty string, there's really no point in rendering
          # a line for this empty field
          continue

        label = self.label_for(field, mapper, name)
        data[name] = (label, value)

      if data:
        panels.append((panel, data,))

    template = filter(bool, (self.options.get('view_template'),
                             'widgets/render_single.html'))

    return Markup(render_template(template,
                                  view=self, related_views=related_views,
                                  csrf_token=csrf.field(),
                                  entity=item, panels=panels, form=form))

  def render_form(self, form, for_new=False, has_save_and_add_new=False):
    # Client-side rules for jQuery.validate
    # See: http://docs.jquery.com/Plugins/Validation/Methods#Validator
    rules = {}
    for field in form:
      rules_for_field = {}
      for validator in field.validators:
        rule_for_validator = getattr(validator, "rule", None)
        if rule_for_validator:
          rules_for_field.update(rule_for_validator)
      if rules_for_field:
        rules[field.name] = rules_for_field
    if rules:
      rules = Markup(json.dumps(rules))
    else:
      rules = None

    template = filter(bool, (self.options.get('edit_template'),
                             'widgets/render_for_edit.html'))

    return Markup(render_template(template,
                                  view=self,
                                  form=form,
                                  for_new=for_new,
                                  has_save_and_add_new=has_save_and_add_new,
                                  rules=rules))

  def label_for(self, field, mapper, name):
    label = field.label
    if label is None:
      try:
        info = mapper.c[name].info
        label = info['label']
      except (AttributeError, KeyError):
        pass

    if label is None:
      try:
        label = _(name)
      except KeyError:
        # i18n may be not initialized (in some unit tests for example)
        label = name

    return label


#
# Used to describe single entity views.
#
class Panel(object):
  """
  `Panel` and `Row` classes help implement a trivial internal DSL for
  specifying multi-column layouts in forms or object views.

  They are currently not really used, since we went with 1-column designs
  eventually.
  """

  def __init__(self, label=None, *rows):
    self.label = label
    self.rows = rows

  def __iter__(self):
    return iter(self.rows)

  def __getitem__(self, item):
    return self.rows[item]

  def __len__(self):
    return len(self.rows)


class Row(object):
  """
  `Panel` and `Row` classes help implement a trivial internal DSL for
  specifying multi-column layouts in forms or object views.

  They are currently not really used, since we went with 1-column designs
  eventually.
  """

  def __init__(self, *cols):
    self.cols = cols

  def __iter__(self):
    return iter(self.cols)

  def __getitem__(self, item):
    return self.cols[item]

  def __len__(self):
    return len(self.cols)


class ModelWidget(object):

  edit_template = 'widgets/model_widget_edit.html'
  view_template = 'widgets/model_widget_view.html'

  def __init__(self, edit_template=None, view_template=None):
    if edit_template is not None:
      self.edit_template = edit_template
    if view_template is not None:
      self.view_template = view_template

  def __call__(self, field, *args, **kwargs):
    return render_template(self.edit_template,
                           form=field,)

  def render_view(self, field, *args, **kwargs):
    _to_skip = (None, False, 0, 0.0, '', u'-')
    rows = []
    for f in field.form:
      if f.is_hidden:
        continue

      value = f.data
      if value in _to_skip and not f.flags.render_empty:
        continue

      value = Markup(f.render_view())
      if value == u'':
        # related models may have [] as value, but we don't discard this type
        # of value in order to let widget a chance to render something useful
        # like an 'add model' button.
        #
        # if it renders an empty string, there's really no point in rendering
        # a line for this empty field
        continue

      label = f.label
      rows.append((label, value))

    return render_template(self.view_template,
                           field=field,
                           rows=rows)


# Form field widgets ###########################################################
class TextInput(wtforms.widgets.TextInput):
  """
  Support pre and post icons.

  An Icon can be a plain string, or an instance of
  :class:`abilian.web.action.Icon`.
  """
  pre_icon = None
  post_icon = None

  def __init__(self, input_type=None, pre_icon=None, post_icon=None,
               *args, **kwargs):
    super(TextInput, self).__init__(input_type, *args, **kwargs)

    if pre_icon is not None:
      self.pre_icon = pre_icon
    if post_icon is not None:
      self.post_icon = post_icon

  def __call__(self, field, *args, **kwargs):
    if not any((self.pre_icon, self.post_icon)):
      return super(TextInput, self).__call__(field, *args, **kwargs)

    kwargs.setdefault('type', self.input_type)
    if 'value' not in kwargs:
      kwargs['value'] = field._value()

    return Markup(render_template_string(
      u'''
      <div class="input-group input-group-type-{{ widget.typename }}">
        {%- if widget.pre_icon %}
          <div class="input-group-addon">{{ widget.pre_icon }}</div>
        {%- endif %}
          <input {{ params | safe}}>
        {%- if widget.post_icon %}
          <div class="input-group-addon">{{ widget.post_icon }}</div>
        {%- endif %}
      </div>
      ''',
      widget=self,
      params=self.html_params(name=field.name, **kwargs)
    ))

  @property
  def typename(self):
    return self.__class__.__name__.lower()


class TextArea(BaseTextArea):
  """
  Accepts "resizeable" parameter: "vertical", "horizontal", "both", None
  """
  _resizeable_valid = ("vertical", "horizontal", "both", None)
  resizeable = None
  rows = None

  def __init__(self, resizeable=None, rows=None, *args, **kwargs):
    BaseTextArea.__init__(self, *args, **kwargs)

    if resizeable not in self._resizeable_valid:
      raise ValueError(
        'Invalid value for resizeable: {}, valid values are: {!r}'
        ''.format(resizeable, self._resizeable_valid)
      )
    if resizeable:
      self.resizeable = 'resizeable-' + resizeable
    else:
      self.resizeable = 'not-resizeable'

    if rows:
      self.rows = int(rows)

  def __call__(self, *args, **kwargs):
    if self.resizeable:
      css = kwargs.get('class_', '')
      kwargs['class_'] = css + ' ' + self.resizeable

    if self.rows and 'rows' not in kwargs:
      kwargs['rows'] = self.rows

    return super(TextArea, self).__call__(*args, **kwargs)


class FileInput(object):
  """
  Renders a file input.  Inspired from
  http://www.abeautifulsite.net/blog/2013/08/whipping-file-inputs-into-shape-with-bootstrap-3/
  """
  def __init__(self, template='widgets/file_input.html'):
    self.template = template

  def __call__(self, field, **kwargs):
    kwargs.setdefault('id', field.id)
    kwargs['name'] = field.name
    kwargs['type'] = 'file'
    kwargs['disabled'] = 'disabled' # JS widget will activate it
    input_elem = u'<input {}>'.format(html_params(**kwargs))
    button_label = _(u'Add file') if 'multiple' in kwargs else _(u'Select file')

    existing = self.build_exisiting_files_list(field)
    uploads = self.build_uploads_list(field)

    if not field.multiple and uploads:
      # single file field: exising file replaced by new upload, don't show
      # existing
      existing = []
    else:
      for data in existing:
        # this might be a very big bytes object: in any case it will not be used
        # in widget's template, and during development it makes very large pages
        # due to debugtoolbar capturing template parameters
        del data['file']

    return Markup(
      render_template(self.template,
                      id=field.id,
                      field=field,
                      widget=self,
                      input=input_elem,
                      button_label=button_label,
                      existing=existing,
                      uploaded=uploads,)
    )

  def build_exisiting_files_list(self, field):
    existing = []
    blob = None
    object_data = field.object_data

    if not object_data:
      return existing

    if not field.multiple:
      object_data = [object_data]
      blob = getattr(field, 'blob')

    for idx, data in enumerate(object_data):
      if data is not None:
        existing.append(
          {
            'file': data,
            'size': len(bytes(data)),
           'delete': idx in field.delete_files_index,
         })
        if hasattr(data, 'filename'):
          existing[-1]['filename'] = data.filename
        elif blob and isinstance(blob, Blob):
          existing[-1]['filename'] = blob.meta.get('filename', u'')

    return existing

  def build_uploads_list(self, field):
    uploads = current_app.extensions['uploads']
    uploaded = []

    for handle in field.upload_handles:
      file_ = uploads.get_file(current_user, handle)
      if file_ is None:
        continue
      meta = uploads.get_metadata(current_user, handle)
      uploaded.append({
        'file': file_,
        'handle': handle,
        'filename': meta.get('filename', handle),
        'size': file_.stat().st_size,
      })

    return uploaded


class ImageInput(FileInput):
  """
  An image widget with client-side preview. To show current image field
  data has to provide an attribute named `url`.
  """
  def __init__(self, template='widgets/image_input.html',
               width=120, height=120, resize_mode=image.CROP,
               valid_extensions=('jpg', 'jpeg', 'png'),):
    super(ImageInput, self).__init__(template=template)
    self.resize_mode = resize_mode
    self.valid_extensions = valid_extensions
    self.width, self.height = width, height

  def build_exisiting_files_list(self, field):
    existing = super(ImageInput, self).build_exisiting_files_list(field)
    for data in existing:
      value = data['file']
      if value:
        if hasattr(value, 'url'):
          image_url = value.url
        else:
          image_url = self.get_b64_thumb_url(
            self.get_thumb(value, self.width, self.height))

        data['image_url'] = image_url

    return existing

  def build_uploads_list(self, field):
    uploaded = super(ImageInput, self).build_uploads_list(field)
    for data in uploaded:
      value = data['file']
      if value:
        if hasattr(value, 'url'):
          image_url = value.url
        else:
          with value.open('rb') as in_:
            image_url = self.get_b64_thumb_url(
              self.get_thumb(in_, self.width, self.height))

        data['image_url'] = image_url

    return uploaded

  def get_thumb(self, data, width, height):
    try:
      get_format(data)
    except IOError:
      return u''
    return image.resize(data, width, height, mode=self.resize_mode)

  def get_b64_thumb_url(self, img):
    try:
      fmt = image.get_format(img).lower()
    except IOError:
      return u''
    thumb = base64.b64encode(img)
    return u'data:image/{format};base64,{img}'.format(format=fmt, img=thumb)

  def render_view(self, field, **kwargs):
    data = field.data
    if not data:
      return u''
    try:
      get_format(data)
    except IOError:
      return u''

    width = kwargs.get('width', self.width)
    height = kwargs.get('heigth', self.height)

    thumb = self.get_thumb(data, width, height)
    width, height = image.get_size(thumb)

    tmpl = u'<img src="{{ url }}" width="{{ width }}" height="{{ height }}" />'
    return render_template_string(tmpl,
                                  url=self.get_b64_thumb_url(thumb),
                                  width=width, height=height)


class Chosen(Select):
  """
  Extends the Select widget using the Chosen jQuery plugin.
  """

  def __call__(self, field, **kwargs):
    kwargs.setdefault('id', field.id)
    html = [u'<select %s class="chzn-select">' % html_params(name=field.name, **kwargs)]
    for val, label, selected in field.iter_choices():
      html.append(self.render_option(val, label, selected))
    html.append(u'</select>')
    return HTMLString(u''.join(html))

  @classmethod
  def render_option(cls, value, label, selected, **kwargs):
    options = dict(kwargs, value=value)
    if selected:
      options['selected'] = True
    return HTMLString(u'<option %s>%s</option>' % (html_params(**options), cgi.escape(unicode(label))))


class TagInput(Input):
  """
  Extends the Select widget using the Chosen jQuery plugin.
  """

  def __call__(self, field, **kwargs):
    kwargs.setdefault('id', field.id)
    kwargs['class_'] = "tagbox"
    if 'value' not in kwargs:
      kwargs['value'] = field._value()

    return HTMLString(u'<input %s>' % self.html_params(name=field.name, **kwargs))


class DateInput(Input):
  """
  Renders date inputs using the fancy Bootstrap Datepicker:
  https://github.com/eternicode/bootstrap-datepicker
  """
  input_type = 'date'

  def __call__(self, field, **kwargs):
    kwargs.setdefault('id', field.id)
    field_id = kwargs.pop('id')
    kwargs.setdefault('name', field.name)
    field_name = kwargs.pop('name')
    kwargs.pop('type', None)
    value = kwargs.pop('value', None)
    if value is None:
      value = field._value()
    if not value:
      value = ''

    date_fmt = kwargs.pop('format', None)
    if date_fmt is not None:
      date_fmt = date_fmt.replace("%", "")\
        .replace("d", "dd")\
        .replace("m", "mm")\
        .replace("Y", "yyyy")
    else:
      date_fmt = get_locale().date_formats['short'].pattern
      date_fmt = babel2datepicker(date_fmt)
      date_fmt = date_fmt.replace('M', 'm')  # force numerical months

    attributes = {
        'class': "input-group date",
        'data-provide': 'datepicker',
        'data-date': value,
        'data-date-format': date_fmt,
        'data-date-autoclose': 'true',
    }

    s = u'<div {}>\n'.format(html_params(**attributes))

    s += u'  <input size="13" type="text" class="form-control" {} />\n'.format(
        html_params(name=field_name, id=field_id, value=value, **kwargs))
    s += u'  <span class="input-group-addon"><i class="fa fa-calendar"></i></span>\n'
    s += u'</div>\n'
    return Markup(s)

  def render_view(self, field, **kwargs):
    return format_date(field.data) if field.data else u''


class TimeInput(Input):
  """
  Renders time inputs using boostrap timepicker:
  https://github.com/jdewit/bootstrap-timepicker
  """
  template = 'widgets/timepicker.html'

  def __init__(self, template=None, widget_mode='dropdown', h24_mode=True, minuteStep=1,
               showSeconds=False, secondStep=1, showInputs=False,
               disableFocus=False, modalBackdrop=False):
    Input.__init__(self)

    if template is not None:
      self.template = template

    self.widget_mode = widget_mode
    self.h24_mode = h24_mode
    self.minuteStep = minuteStep
    self.showSeconds = showSeconds
    self.secondStep = secondStep
    self.showInputs = showInputs
    self.disableFocus = disableFocus
    self.modalBackdrop = modalBackdrop
    self.strptime = u'%H:%M' if self.h24_mode else u'%I:%M %p'

  def __call__(self, field, **kwargs):
    kwargs.setdefault('id', field.id)
    field_id = kwargs.pop('id')
    value = kwargs.pop('value', None)
    if value is None:
      value = field._value()
    if not value:
      value = ''

    time_fmt = get_locale().time_formats['short'].format
    is_h12 = ('%(h)s' in time_fmt or '%(K)s' in time_fmt)

    input_params = {
      'data-template': self.widget_mode,
      'data-show-meridian': is_h12,
      'data-minute-step': self.minuteStep,
      'data-show-seconds': self.showSeconds,
      'data-second-step': self.secondStep,
      'data-show-inputs': self.showInputs,
      'data-disable-focus': self.disableFocus,
      'data-modal-backdrop': self.modalBackdrop
    }

    input_params = {k: Markup(json.dumps(v)) for k, v in input_params.items()}

    return Markup(render_template(self.template,
                                  id=field_id, value=value,
                                  field=field,
                                  required=False,
                                  timepicker_attributes=input_params))


class DateTimeInput(object):
  """
  if corresponding field.raw_data exist it is used
  to initialize default date & time (raw_data example: ["10/10/16 | 09:00"])
  """

  def __init__(self):
    self.date = DateInput()
    self.time = TimeInput()

  def __call__(self, field, **kwargs):
    kwargs.setdefault('id', field.id)
    field_id = kwargs.pop('id')
    kwargs.setdefault('name', field.name)
    field_name = kwargs.pop('name')

    locale = get_locale()
    date_fmt = locale.date_formats['short'].pattern
    date_fmt = babel2datetime(date_fmt)
    date_fmt = date_fmt.replace('%B', '%m').replace('%b', '%m')  # force numerical months
    time_fmt = u'%H:%M'

    value = kwargs.pop('value', None)
    if value is None:
      if field.data:
        value = field.data
        date_value = value.strftime(date_fmt) if value else u''
        time_value = value.strftime(time_fmt) if value else u''
      elif field.raw_data:
        value = field.raw_data  # example "10/10/16 | 09:00"
        value = u''.join(value)
        value = value.split(u'|')

        try:
          date_value = datetime.strptime(value[0].strip(), date_fmt)
        except (ValueError, IndexError):
          date_value = u''

        try:
          time_value = value[1].strip()  # example  "09:00"
        except IndexError:
          time_value = u''

      else:
        date_value = u''
        time_value = u''

    return (
      Markup(
        u'<div class="form-inline">\n'
        u'<input class="datetimepicker" type="hidden" id="{id}" name="{name}" '
        u'value="{date} | {time}" />\n'
        u''.format(id=field_id, name=field_name, date=date_value, time=time_value))
      +
      self.date(field,
                id=field_id + '-date', name=field_name + '-date',
                value=date_value)
      +
      self.time(field,
                  id=field_id + '-time', name=field_name + '-time',
                  value=time_value)
      + Markup(u'</div>')
    )


class DefaultViewWidget(object):
  def render_view(self, field, **kwargs):
    value = field.object_data
    if isinstance(value, string_types):
      return text2html(value)
    else:
      return unicode(value or u'')  # [], None and other must be rendered using
                                    # empty string


class BooleanWidget(wtforms.widgets.CheckboxInput):

  # valid data-* options when using boostrap-switch
  _ON_OFF_VALID_OPTIONS = frozenset((
    'animate', 'indeterminate', 'inverse', 'radio-all-off',
    'on-color', 'off-color', 'on-text', 'off-text', 'label-text',
    'handle-width', 'label-width', 'base-class', 'wrapper-class'
  ))

  def __init__(self, *args, **kwargs):
    self.on_off_mode = kwargs.pop('on_off_mode', False)
    self.on_off_options = {}
    on_off_options = kwargs.pop('on_off_options', {})
    for k, v in on_off_options.iteritems():
      if k not in self._ON_OFF_VALID_OPTIONS:
        continue
      self.on_off_options['data-' + k] = v

    if self.on_off_mode:
      self.on_off_options['data-toggle'] = u'on-off'

    super(BooleanWidget, self).__init__(*args, **kwargs)

  def __call__(self, field, **kwargs):
    if self.on_off_mode:
      kwargs.update(self.on_off_options)

    return super(BooleanWidget, self).__call__(field, **kwargs)

  def render_view(self, field, **kwargs):
    return u'\u2713' if field.object_data else u''  # Unicode "Check mark"


class PasswordInput(BasePasswordInput):
  """
  Supports setting 'autocomplete' at instanciation time
  """
  def __init__(self, *args, **kwargs):
    self.autocomplete = kwargs.pop('autocomplete', None)
    BasePasswordInput.__init__(self, *args, **kwargs)

  def __call__(self, field, **kwargs):
    kwargs.setdefault('autocomplete', self.autocomplete)
    return BasePasswordInput.__call__(self, field, **kwargs)

  def render_view(self, field, **kwargs):
    return u'*****'


class FloatWidget(wtforms.widgets.TextInput):
  """ In view mode, format float number to 'precision' decimal
  """
  def __init__(self, precision=None):
    self.precision = precision
    if precision is not None:
        self._fmt = '.{:d}f'.format(precision)

  def render_view(self, field, **kwargs):
    data = field.object_data
    if data is None:
      return u''

    return format(data, self._fmt)


class DateWidget(wtforms.widgets.TextInput):
  def render_view(self, field, **kwargs):
    return (format_date(field.object_data)
            if field.object_data else u'')


class DateTimeWidget(DateWidget):
  def render_view(self, field, **kwargs):
    return (format_datetime(field.object_data)
            if field.object_data else u'')


class EntityWidget(object):
  def render_view(self, field, **kwargs):
    objs = field.object_data
    if not field.multiple:
      objs = [objs]
    return u', '.join(
      u'<a href="{}">{}</a>'.format(url_for(o), cgi.escape(o.name))
      for o in objs
      if o)


class HoursWidget(TextInput):
  """ Widget used to show / enter hours.
  Currently hardcoded to heure(s)
  """
  post_icon = _l(u'hour(s)')
  input_type = 'number'

  def render_view(self, field, **kwargs):
    val = field.object_data
    unit = u'h'

    if val is None:
      return u''

    # \u00A0: non-breakable whitespace
    return u'{value}\u00A0{unit}'.format(value=val, unit=unit)


class MoneyWidget(TextInput):
  """
  Widget used to show / enter money amount.

  Currently hardcoded to € / k€.
  """
  post_icon = u'€'
  input_type = 'number'

  def render_view(self, field, **kwargs):
    val = field.object_data
    unit = u'€'

    if val is None:
      return u''

    if val > 1000:
      unit = u'k€'
      val = int(round(val / 1000.0))

    # `format_currency()` is not used since it display numbers with cents
    #units, which we don't want
    #
    # \u00A0: non-breakable whitespace
    return u'{value}\u00A0{unit}'.format(value=format_number(val), unit=unit)


class EmailWidget(TextInput):
  pre_icon = u'@'

  def render_view(self, field, **kwargs):
    links = u''
    if isinstance(field, wtforms.fields.FieldList):
      for entry in field.entries:
        link = bleach.linkify(entry.data, parse_email=True)
        if link:
          links = links + u' {}&nbsp;<i class="fa fa-envelope"></i><br>'.format(link)
    else:
      link = bleach.linkify(field.object_data, parse_email=True)
      if link:
        links = u'{}&nbsp;<i class="fa fa-envelope"></i>'.format(link)
    return links


class URLWidget(object):
  def render_view(self, field, **kwargs):
    return (linkify_url(field.object_data)
            if field.object_data else u'')


class RichTextWidget(object):
  template = 'widgets/richtext.html'
  allowed_tags = {
      'a': {'href': True, 'title': True},
      'abbr': {'title': True},
      'acronym': {'title': True},
      'b': True,
      'blockquote': True,
      'br': True,
      'code': True,
      'em': True,
      'h1': True, 'h2': True, 'h3': True, 'h4': True, 'h5': True, 'h6': True,
      'i': True,
      'img': {'src': True},
      'li': True,
      'ol': True,
      'strong': True,
      'ul': True,
      'p': True,
      'u': True
  }

  def __init__(self, allowed_tags=None, template=None):
    if allowed_tags is not None:
      self.allowed_tags = allowed_tags
    if template is not None:
      self.template = template

  def __call__(self, field, **kwargs):
    value = kwargs.pop('value') if 'value' in kwargs else field._value()
    kwargs.setdefault('allowed_tags', self.allowed_tags)
    return render_template(self.template, field=field, value=value, kw=kwargs)


class ListWidget(wtforms.widgets.ListWidget):
  """ display field label is optionnal
  """

  def __init__(self, html_tag='ul', prefix_label=True, show_label=True):
    wtforms.widgets.ListWidget.__init__(self, html_tag, prefix_label)
    self.show_label = show_label

  def __call__(self, field, **kwargs):
    if self.show_label:
      return super(ListWidget, self).__call__(field, **kwargs)

    kwargs.setdefault('id', field.id)
    html = [u'<%s %s>' % (self.html_tag, wtforms.widgets.html_params(**kwargs))]
    for subfield in field:
      html.append(u'<li>{}</li>'.format(subfield()))

    html.append(u'</%s>' % self.html_tag)
    return wtforms.widgets.HTMLString(''.join(html))

  def render_view(self, field, **kwargs):
    data = field.data
    is_empty = data == [] if field.multiple else data is None

    if not is_empty:
      data = ([label for v, label, checked in field.iter_choices() if checked]
              if hasattr(field, 'iter_choices') and callable(field.iter_choices)
              else field.object_data)
    else:
      data = []

    return Markup(render_template_string(
      '''{%- for obj in data %}
      <span class="badge">{{ obj }}</span>
      {%- endfor %}''',
      data=data))


class FieldListWidget(object):
  """
  For list of Fields
  (using <tr><td>)
  """
  view_template = 'widgets/fieldlist_view.html'
  template = 'widgets/fieldlist.html'

  def __init__(self, template=None, view_template=None):
    if template is not None:
      self.template = template
    if view_template is not None:
      self.view_template = view_template

  def __call__(self, field, **kwargs):
    assert isinstance(field, wtforms.fields.FieldList)
    return Markup(render_template(self.template, field=field))

  def render_view(self, field, **kwargs):
    assert isinstance(field, wtforms.fields.FieldList)
    value = field.object_data
    if not value:
      return ''
    return Markup(render_template(self.view_template, field=field))


class TabularFieldListWidget(object):
  """
  For list of formfields

  2 templates are available:
   * widgets/tabular_fieldlist_widget.html (default):
     Show sub-forms as a table, one row of inputs per model
   * widgets/model_fieldlist.html:
     Show sub-forms as a list of forms
  """

  def __init__(self, template='widgets/tabular_fieldlist_widget.html'):
    self.template = template

  def __call__(self, field, **kwargs):
    assert isinstance(field, wtforms.fields.FieldList)
    labels = None

    if len(field):
      assert isinstance(field[0], wtforms.fields.FormField)
      field_names = [f.short_name for f in field[0] if not f.is_hidden]
      data_type = field.entries[0].__class__.__name__ + 'Data'
      Data = namedtuple(data_type, field_names)
      labels = Data(*[f.label for f in field[0] if not f.is_hidden])

    return Markup(render_template(self.template, labels=labels, field=field))


class ModelListWidget(object):

  def __init__(self, template='widgets/horizontal_table.html'):
    self.template = template

  def render_view(self, field, **kwargs):
    assert isinstance(field, ModelFieldList)
    value = field.object_data
    if not value:
      return render_template(self.template, field=field, labels=(),
                             rows=(), **kwargs)

    field_names = field._field_names
    labels = field._field_labels
    data_type = field.entries[0].object_data.__class__.__name__ + 'Data'
    Data = namedtuple(data_type, field_names)
    labels = Data(*labels)

    rows = []
    for entry in field.entries:
      row = []
      for f in ifilter(lambda f: not f.is_hidden, entry.form):
        row.append(Markup(f.render_view()))

      rows.append(Data(*row))

    rendered = render_template(self.template, field=field, labels=labels,
                               rows=rows, **kwargs)
    return rendered


#
# Selection widget.
#
class Select2(Select):
  """
  Transforms a Select widget into a Select2 widget. Depends on global JS code.
  """
  unescape_html = False

  def __init__(self, unescape_html=False, js_init='select2', *args, **kwargs):
    super(Select2, self).__init__(*args, **kwargs)
    self.js_init = js_init

    if unescape_html:
      self.unescape_html = True

  def __call__(self, field, *args, **kwargs):
    # 'placeholder' option presence is required for 'allowClear'
    params = {'placeholder': u''}
    if self.unescape_html:
      params['makeHtml'] = True
    if not field.flags.required:
      params['allowClear'] = True

    css_class = kwargs.setdefault('class_', u'')
    if 'js-widget' not in css_class:
      css_class += u' js-widget'
      kwargs['class_'] = css_class

    kwargs.setdefault('data-init-with', self.js_init)
    kwargs['data-init-params'] = json.dumps(params)
    return Select.__call__(self, field, *args, **kwargs)

  def render_view(self, field, **kwargs):
    labels = [unicode(label)
              for v, label, checked in field.iter_choices()
              if checked]
    return u'; '.join(labels)

  @classmethod
  def render_option(cls, value, label, selected, **kwargs):
    if value is None:
      return HTMLString('<option></option>')
    return Select.render_option(value, label, selected, **kwargs)


class Select2Ajax(object):
  """
  Ad-hoc select widget based on Select2.

  The code below is probably very fragile, since it depends on the internal
  structure of a Select2 widget.

  :param format_result: `formatResult` arg of Select2. Must be a valid
  javascript reference (like `Abilian.my_format_function`)

  :param format_selection: like `format_result`, but for Select2's
  `formatSelection`

  :param values_builder: an optional function, which given a data iterable of
  values, returns a list of dict items suitable for widget initial data;
  generally needed when using custom format_result/format_selection.
  """
  def __init__(self, template='widgets/select2ajax.html', multiple=False,
               format_result=None, format_selection=None,
               values_builder=None):
    self.template = template
    self.multiple = multiple
    self.values_builder = (
      values_builder
      if callable(values_builder)
      else lambda data: [{'id': o.id, 'text': o.name} for o in data if o]
    )
    self.s2_params = dict(multiple=self.multiple)

    if format_result:
      self.s2_params['formatResult'] = format_result
    if format_selection:
      self.s2_params['formatSelection'] = format_selection

  def process_formdata(self, valuelist):
    """
    field helper: as of select2 3.x, multiple values are passed as a single
    string.
    """
    if self.multiple:
      valuelist = valuelist[0].split(u',')
    return valuelist

  def __call__(self, field, **kwargs):
    """
    Render widget
    """
    if self.multiple:
      kwargs['multiple'] = True

    css_class = kwargs.setdefault('class_', u'')
    if 'class' in kwargs:
      css_class += kwargs.pop('class', u'')
      kwargs['class_'] = css_class

    if 'js-widget' not in css_class:
      css_class += u' js-widget'
      kwargs['class_'] = css_class

    s2_params = {}
    s2_params.update(self.s2_params)

    if field.ajax_source:
      s2_params['ajax'] = {'url': unicode(field.ajax_source)}

    s2_params['placeholder'] = unicode(field.label.text)

    if not field.flags.required:
      s2_params['allowClear'] = True

    data = field.data
    if not self.multiple:
      data = [data]

    values = self.values_builder(data)
    input_value = u','.join(unicode(o.id) for o in data if o)
    data_node_id = None

    json_data = {}
    if values:
      json_data['values'] = values
      data_node_id = s2_params['dataNodeId'] = field.id + '-json'

    kwargs['data-init-with'] = 'select2ajax'
    kwargs['data-init-params'] = json.dumps(s2_params)


    extra_args = Markup(html_params(**kwargs))

    return Markup(render_template(self.template,
                                  field=field,
                                  name=field.name,
                                  id=field.id,
                                  input_value=input_value,
                                  json_data=json_data,
                                  required=not field.allow_blank,
                                  data_node_id=data_node_id,
                                  extra_args=extra_args,))



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import sqlalchemy as sa
from flask import current_app, send_file
from werkzeug.exceptions import BadRequest
from werkzeug.utils import redirect

from abilian.core.entities import Entity
from abilian.core.models.attachment import Attachment, is_support_attachments
from abilian.i18n import _, _l
from abilian.web import nav, url_for
from abilian.web.action import ButtonAction, actions
from abilian.web.blueprints import Blueprint
from abilian.web.views import (BaseObjectView, ObjectCreate, ObjectDelete,
                               ObjectEdit)

from .forms import AttachmentForm

bp = Blueprint('attachments', __name__, url_prefix='/attachments')


def _default_attachment_view(obj, obj_type, obj_id, **kwargs):
  if not hasattr(obj, 'entity'):
    return url_for('attachments.entity', object_id=obj_id)
  entity = obj.entity
  return url_for(entity, _anchor='attachment-{}'.format(obj.id))


@bp.record_once
def register_default_view(state):
  state.app.default_view.register(Attachment, _default_attachment_view)

UPLOAD_BUTTON = ButtonAction('form', 'edit', btn_class='primary',
                             title=_l(u'Send'))


class BaseAttachmentView(object):
  """
  Mixin for attachment views
  """
  Model = Attachment
  Form = AttachmentForm

  #: owning entity
  entity = None

  def init_object(self, args, kwargs):
    args, kwargs = super(BaseAttachmentView, self).init_object(args, kwargs)
    entity_id = kwargs.pop('entity_id', None)

    if entity_id is not None:
      self.entity = Entity.query.get(entity_id)

    if self.entity is None:
      raise BadRequest('No entity provided')

    if not is_support_attachments(self.entity):
      raise BadRequest('This entity is doesn\'t support attachments')

    extension = current_app.extensions['attachments']
    self.Form = extension.manager(self.entity).Form
    actions.context['object'] = self.entity
    return args, kwargs

  def view_url(self):
    kw = {}
    if self.obj and self.obj.id:
      kw['_anchor'] = 'attachment-{}'.format(self.obj.id)
    return url_for(self.entity, **kw)

  def index_url(self):
    return self.view_url()

  @property
  def activity_target(self):
    return self.entity


class AttachmentDownload(BaseAttachmentView, BaseObjectView):
  """
  """
  def get(self):
    blob = self.obj.blob
    metadata = blob.meta
    filename = metadata.get('filename', self.obj.name)
    content_type = metadata.get('mimetype')
    stream = blob.file.open('rb')

    return send_file(stream,
                     as_attachment=True,
                     attachment_filename=filename,
                     mimetype=content_type,
                     cache_timeout=0,
                     add_etags=False,
                     conditional=False)


download_view = AttachmentDownload.as_view('download')
bp.route('/<int:entity_id>/<int:object_id>/download')(download_view)


class AttachmentEdit(BaseAttachmentView, ObjectEdit):
  _message_success = _l(u'Attachment edited')


edit_view = AttachmentEdit.as_view('edit')
bp.route('/<int:entity_id>/<int:object_id>/edit')(edit_view)


class AttachmentCreateView(BaseAttachmentView, ObjectCreate):
  """
  """
  _message_success = _l(u'Attachment added')

  def init_object(self, args, kwargs):
    args, kwargs = super(AttachmentCreateView, self).init_object(args, kwargs)

    self.obj.entity = self.entity
    session = sa.orm.object_session(self.entity)

    if session:
      sa.orm.session.make_transient(self.obj)

    return args, kwargs

  def breadcrumb(self):
    label = _(u'New attachment on "{title}"').format(title=self.entity.name)
    return nav.BreadcrumbItem(label=label)

  def get_form_buttons(self, *args, **kwargs):
    return [UPLOAD_BUTTON]


create_view = AttachmentCreateView.as_view('create')
bp.route('/<int:entity_id>/create')(create_view)


class AttachmentDelete(BaseAttachmentView, ObjectDelete):
  pass

delete_view = AttachmentDelete.as_view('delete')
bp.route('/<int:entity_id>/<int:object_id>/delete')(delete_view)


class AttachmentEntity(BaseObjectView):
  """
  Redirects to an attachment's entity view
  """
  Model = Attachment

  def get(self):
    return redirect(url_for(self.obj))


entity_view = AttachmentEntity.as_view('entity')
bp.route('/<int:object_id>/entity')(entity_view)



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import sys

from blinker import Namespace, Signal
from flask import current_app, g
from flask_debugtoolbar.panels import DebugPanel

from abilian.web.action import actions


class ActionDebugPanel(DebugPanel):
  name = 'Actions'

  user_enable = True
  has_content = True

  def nav_title(self):
    return 'Actions'

  def title(self):
    return 'Actions'

  def url(self):
    return ''

  def content(self):
    actions_for_template = []

    for category in actions.actions().keys():
      available_actions = actions.for_category(category)
      for action in available_actions:
        d = {
          'category': action.category,
          'title': action.title,
          'class': action.__class__.__name__,
        }
        try:
          d['endpoint'] = unicode(action.endpoint)
        except:
          d['endpoint'] = '<Exception>'
        try:
          d['url'] = unicode(action.url(g.action_context))
        except:
          d['url'] = '<Exception>'
        actions_for_template.append(d)

    actions_for_template.sort(key=lambda x: (x['category'], x['title']))

    ctx = {'actions': actions_for_template}

    jinja_env = current_app.jinja_env
    jinja_env.filters.update(self.jinja_env.filters)
    template = jinja_env.get_or_select_template(
      'debug_panels/actions_panel.html')
    return template.render(ctx)


class SignalsDebugPanel(DebugPanel):
  name = 'Signals'

  user_enable = True
  has_content = True

  events = []

  def nav_title(self):
    return 'Signals'

  def title(self):
    return 'Signals'

  def url(self):
    return ''

  def content(self):
    module_names = sys.modules.keys()
    module_names.sort()

    signals = []

    for module_name in module_names:
      module = sys.modules[module_name]
      if not module:
        continue
      module_vars = vars(module)
      for var_name, var in module_vars.items():
        if not isinstance(var, Namespace):
          continue

        ns = var
        ns_name = var_name
        for signal_name, signal in ns.items():
          d = {
            'module_name': module_name,
            'ns_name': ns_name,
            'signal_name': signal_name,
            'signal': signal,
            'receivers': [unicode(r) for r in signal.receivers.values()]
          }
          signals.append(d)

    signals.sort(
      key=lambda x: (d['module_name'], d['ns_name'], d['signal_name']))

    ctx = {'signals': signals, 'events': self.events}

    jinja_env = current_app.jinja_env
    jinja_env.filters.update(self.jinja_env.filters)
    template = jinja_env.get_or_select_template(
      'debug_panels/signals_panel.html')
    return template.render(ctx)

  def process_request(self, request):
    self.events = []

    if getattr(Signal.send, '__wrapped', False):
      return

    orig_send = Signal.send
    events = self.events

    def wrapped_send(self, *sender, **kwargs):
      d = {
        'signal_name': self.name,
        'sender': unicode(sender[0]),
        'args': kwargs,
      }
      events.append(d)
      return orig_send(self, *sender, **kwargs)

    wrapped_send.__wrapped = True
    Signal.send = wrapped_send



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import imghdr
from cStringIO import StringIO

import babel
import PIL.Image
from flask import (current_app, flash, g, redirect, render_template, request,
                   url_for)
from werkzeug.exceptions import InternalServerError
from wtforms.fields import StringField
from wtforms.validators import ValidationError

from abilian.i18n import _, _l, get_default_locale
from abilian.services.preferences.panel import PreferencePanel
from abilian.web import csrf
from abilian.web.forms import Form, fields, validators, widgets


class UserPreferencesForm(Form):

  password = StringField(_l(u'New Password'),
                         widget=widgets.PasswordInput(autocomplete='off'))
  confirm_password= StringField(_l(u'Confirm new password'),
                                widget=widgets.PasswordInput(autocomplete='off'))

  photo = fields.FileField(label=_l('Photo'),
                           widget=widgets.ImageInput(width=55, height=55))

  locale = fields.LocaleSelectField(
    label=_l(u'Preferred Language'),
    validators=(validators.required(),),
    default=lambda: get_default_locale(),
  )

  timezone = fields.TimezoneField(
    label=_l(u'Time zone'),
    validators=(validators.required(),),
    default=babel.dates.LOCALTZ,
  )

  def validate_password(self, field):
    pwd = field.data
    confirmed = self['confirm_password'].data

    if pwd != confirmed:
      raise ValidationError(
        _(u'Passwords differ. Ensure you have typed same password in both'
          u' "password" field and "confirm password" field.'))

  def validate_photo(self, field):
    data = request.form.get(field.name)
    if not data:
      return

    data = field.data
    filename = data.filename
    valid = any(filename.lower().endswith(ext)
                for ext in ('.png', '.jpg', '.jpeg'))

    if not valid:
      raise ValidationError(_(u'Only PNG or JPG image files are accepted'))

    img_type = imghdr.what('ignored', data.read())

    if not img_type in ('png', 'jpeg'):
      raise ValidationError(_(u'Only PNG or JPG image files are accepted'))

    data.seek(0)
    try:
      # check this is actually an image file
      im = PIL.Image.open(data)
      im.load()
    except:
      raise ValidationError(_(u'Could not decode image file'))

    # convert to jpeg
    #FIXME: better do this at model level?
    jpeg = StringIO()
    im.convert('RGBA').save(jpeg, 'JPEG')
    field.data = jpeg.getvalue()


class UserPreferencesPanel(PreferencePanel):
  id = 'user'
  label = _l(u'About me')

  def is_accessible(self):
    return True

  def get(self):
    # Manual security check, should be done by the framework instead.
    if not self.is_accessible():
      raise InternalServerError()

    data = {}
    photo = g.user.photo
    if photo:
      # subclass str/bytes to set additional 'url' attribute
      photo = type('Photo', (bytes,),
                   dict(object=photo,
                        url=url_for('users.photo', user_id=g.user.id)))
      data['photo'] = photo

    form = UserPreferencesForm(obj=g.user, formdata=None, prefix=self.id, **data)
    if form['locale'].data is None:
      form['locale'].data = get_default_locale()
    return render_template('preferences/user.html', form=form, title=self.label)

  @csrf.support_graceful_failure
  def post(self):
    # Manual security check, should be done by the framework instead.
    if not self.is_accessible():
      raise InternalServerError()

    if request.form['_action'] == 'cancel':
      return redirect(url_for('.user'))

    form = UserPreferencesForm(request.form, prefix=self.id)

    if form.validate():
      if request.csrf_failed:
        current_app.extensions['csrf-handler'].flash_csrf_failed_message()
        return render_template('preferences/user.html', form=form)

      del form.confirm_password
      if form.password.data:
        g.user.set_password(form.password.data)
        flash(_(u'Password changed'), 'success')
      del form.password

      form.populate_obj(g.user)

      current_app.db.session.commit()
      flash(_(u"Preferences saved."), "info")
      return redirect(url_for(".user"))
    else:
      return render_template('preferences/user.html', form=form)



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import abc
from functools import total_ordering

import sqlalchemy as sa

from abilian.core.entities import Entity

from .base import IdMixin, Model

#: backref attribute on tagged elements
TAGS_ATTR = '__tags__'


class SupportTagging(object):
  """
  """
  __metaclass__ = abc.ABCMeta


def register(cls):
  """
  Register an :class:`~.Entity` as a taggable class.

  Can be used as a class decorator:

  .. code-block:: python

      @tag.register
      class MyContent(Entity):
          ....
  """
  if not issubclass(cls, Entity):
    raise ValueError('Class must be a subclass of abilian.core.entities.Entity')

  SupportTagging.register(cls)
  return cls


def is_support_tagging(obj):
  """
  :param obj: a class or instance
  """
  if isinstance(obj, type):
    return issubclass(obj, SupportTagging)

  if not isinstance(obj, SupportTagging):
    return False

  if obj.id is None:
    return False

  return True


entity_tag_tbl = sa.Table(
  'entity_tags', Model.metadata,
  sa.Column('tag_id', sa.Integer, sa.ForeignKey('tag.id', ondelete='CASCADE')),
  sa.Column('entity_id', sa.Integer, sa.ForeignKey(Entity.id, ondelete='CASCADE')),
  sa.UniqueConstraint('tag_id', 'entity_id'),
)


@total_ordering
class Tag(IdMixin, Model):
  """
  Tags are text labels that can be attached to :class:`entities <.Entity>`.

  They are namespaced, so that independent group of tags can be defined in the
  application. The default namespace is `"default"`.
  """
  __tablename__ = 'tag'

  #: namespace
  ns = sa.Column(sa.UnicodeText(), nullable=False,
                 default=u'default', server_default='default')

  #: Label visible to the user
  label = sa.Column(sa.UnicodeText(), nullable=False)

  #: :class:`entities <.Entity>` attached to this tag
  entities = sa.orm.relationship(
    Entity, collection_class=set,
    secondary=entity_tag_tbl,
    backref=sa.orm.backref(TAGS_ATTR, collection_class=set),
  )

  __mapper_args__ = {
    'order_by': label,
  }

  __table_args__ = (
    sa.UniqueConstraint(ns, label),
    # namespace is not empty and is not surrounded by space characters
    sa.CheckConstraint(
      sa.sql.and_(
        sa.sql.func.trim(ns) == ns,
        ns != u''),
    ),
    # label is not empty and is not surrounded by space characters
    sa.CheckConstraint(
      sa.sql.and_(
        sa.sql.func.trim(label) == label,
        label != u''),
    ),
  )

  def __unicode__(self):
    return self.label

  def __lt__(self, other):
    return unicode(self).lower().__lt__(unicode(other).lower())

  def __repr__(self):
    cls = self.__class__
    return '<{mod}.{cls} id={t.id!r} ns={t.ns!r} label={t.label!r} at 0x{addr:x}>'.format(
      mod=cls.__module__, cls=cls.__name__, t=self, addr=id(self),
    )



# coding=utf-8
"""
Base classes for file download
"""
from __future__ import absolute_import

from datetime import datetime, timedelta

from flask import request, send_file
from werkzeug.exceptions import BadRequest

from abilian.core.util import utc_dt

from .base import View


class BaseFileDownload(View):
  """
  """
  set_expire = False
  expire_offset = timedelta(days=365)
  as_attachment = False

  #: argument name that must be found in view kwargs. This is a safety measure
  #: to prevent setting far expire date on resources without a varying argument
  #: in url (path or query string), such as a timestamp, a serial, a hash...
  expire_vary_arg = None

  def __init__(self, set_expire=None, expire_offset=None,
               expire_vary_arg=None, as_attachment=None):
    # override class default value only if arg is specified in constructor. This
    # allows subclasses to easily override theses defaults.
    if set_expire is not None:
      self.set_expire = set_expire
    if expire_offset is not None:
      self.expire_offset = expire_offset
    if expire_vary_arg is not None:
      self.expire_vary_arg = expire_vary_arg
    if as_attachment is not None:
      self.as_attachment = bool(as_attachment)

    if self.set_expire:
      if not self.expire_offset:
        raise ValueError('expire_offset is not set')
      if not self.expire_vary_arg:
        raise ValueError('expire_vary_arg is not set')

  def prepare_args(self, args, kwargs):
    if self.set_expire:
      vary_arg = kwargs.get(self.expire_vary_arg,
                            request.args.get(self.expire_vary_arg))
      if vary_arg is None:
        # argument for timestamp, serial etc is missing. We must refuse to serve
        # an image with expiry date set up to maybe 1 year from now.
        # Check the code that has generated this url!
        raise BadRequest(
          'File version marker is missing ({}=?)'.format(
            repr(self.expire_vary_arg))
        )

    args, kwargs = super(BaseFileDownload, self).prepare_args(args, kwargs)
    kwargs['attach'] = request.args.get('attach', self.as_attachment, type=bool)
    return args, kwargs

  def make_response(self, *args, **kwargs):
    # for example: return flask.make_response(...)
    # or: return flask.send_file(...)
    raise NotImplementedError()

  def get_filename(self, *args, **kwargs):
    raise NotImplementedError()

  def get_content_type(self, *args, **kwargs):
    raise NotImplementedError()

  def get(self, attach, *args, **kwargs):
    """
    :param image: image as bytes
    :param s: requested maximum width/height size
    """
    response = self.make_response(*args, **kwargs)
    response.headers['content-type'] = self.get_content_type(*args, **kwargs)

    if attach:
      filename = self.get_filename(*args, **kwargs)
      if not filename:
        filename = u'file.bin'
      response.headers.add('Content-Disposition', u'attachment',
                           filename=filename)

    self.set_cache_headers(response)
    return response

  def set_cache_headers(self, response):
    if self.set_expire:
      response.cache_control.public = False
      response.cache_control.private = True
      response.cache_control.max_age = int(self.expire_offset.total_seconds())
      response.expires = utc_dt(datetime.utcnow() + self.expire_offset)


class BaseBlobDownload(BaseFileDownload):

  def get_blob(self, *args, **kwargs):
    raise NotImplementedError()

  def prepare_args(self, args, kwargs):
    args, kwargs = super(BaseBlobDownload, self).prepare_args(args, kwargs)
    self.blob = self.get_blob(*args, **kwargs)
    metadata = self.blob.meta
    self.filename = metadata.get('filename', self.obj.name)
    self.content_type = metadata.get('mimetype')
    return args, kwargs

  def get_filename(self, *args, **kwargs):
    return self.filename

  def get_content_type(self, *args, **kwargs):
    return self.content_type

  def make_response(self, *args, **kwargs):
    blob = self.blob
    stream = blob.file.open('rb')

    return send_file(stream,
                     as_attachment=False,
                     mimetype=self.content_type,
                     cache_timeout=0,
                     add_etags=False,
                     conditional=False)



# coding=utf-8
"""
Class based views
"""
from __future__ import absolute_import, division, print_function

import logging

import sqlalchemy as sa
from flask import (current_app, flash, g, redirect, render_template, request,
                   url_for)
from werkzeug.exceptions import BadRequest, NotFound

from abilian.core.entities import ValidationError
from abilian.core.signals import activity
from abilian.i18n import _, _l
from abilian.services.security import CREATE, DELETE, READ, WRITE

from .. import csrf, forms, nav
from ..action import ButtonAction, Endpoint, actions
from .base import JSONView, View

logger = logging.getLogger(__name__)


class BaseObjectView(View):
  """
  Base class common to all database objects views
  """
  #: form title
  title = None

  #: Model class
  Model = None

  #: primary key name to look for in url arguments
  pk = 'object_id'

  #: object instance for this view
  obj = None

  #: object id
  object_id = None

  #: template to render
  template = None

  #: default templates inherit from "base_template". This allows to use generic
  #: templates with a custom base
  base_template = "base.html"

  def __init__(self, Model=None, pk=None, base_template=None, *args, **kwargs):
    View.__init__(self, *args, **kwargs)
    cls = self.__class__
    self.pk = pk if pk is not None else cls.pk
    self.Model = Model if Model is not None else cls.Model
    self.base_template = (base_template
                          if base_template is not None
                          else cls.base_template)

  def prepare_args(self, args, kwargs):
    args, kwargs = self.init_object(args, kwargs)

    if self.obj is None:
      raise NotFound()

    return args, kwargs

  def breadcrumb(self):
    """
    Return :class:`..nav.BreadcrumbItem` instance for this object.

    This method may return a list of BreadcrumbItem instances. Return
    `None` if nothing.
    """
    return None

  def init_object(self, args, kwargs):
    """
    This method is reponsible for setting :attr:`obj`. It is called during
    :meth:`prepare_args`.
    """
    self.object_id = kwargs.pop(self.pk, None)
    if self.object_id is not None:
      self.obj = self.Model.query.get(self.object_id)
      actions.context['object'] = self.obj

    return args, kwargs

  def get(self, *args, **kwargs):
    bc = self.breadcrumb()
    if bc is not None:
      bc = [bc] if isinstance(bc, nav.BreadcrumbItem) else list(bc)
      assert all(isinstance(b, nav.BreadcrumbItem) for b in bc)
      g.breadcrumb.extend(bc)

    kwargs = {'base_template': self.base_template}
    kwargs.update(self.template_kwargs)
    # forbid override "view"
    kwargs['view'] = self
    return render_template(self.template, **kwargs)

  @property
  def template_kwargs(self):
    """
    Template render arguments. You can override `base_template` for
    instance. Only `view` cannot be overriden.
    """
    return {}


class ObjectView(BaseObjectView):
  """
  View objects
  """

  #: html template
  template = 'default/object_view.html'

  #: View form class. Form object used to show objects fields
  Form = None

  #: required permission. Must be an instance of
  #: :class:`abilian.services.security.Permission`
  permission = READ

  #: form instance for this view
  form = None

  def __init__(self, Model=None, pk=None, Form=None, template=None,
               *args, **kwargs):
    super(ObjectView, self).__init__(Model, pk, *args, **kwargs)
    cls = self.__class__
    self.Form = Form if Form is not None else cls.Form
    self.template = template if template is not None else cls.template

  def prepare_args(self, args, kwargs):
    """
    :attr:`form` is initialized here. See also :meth:`View.prepare_args`.
    """
    args, kwargs = super(ObjectView, self).prepare_args(args, kwargs)
    self.form = self.Form(**self.get_form_kwargs())
    return args, kwargs

  def get_form_kwargs(self):
    kw = dict(obj=self.obj)
    if issubclass(self.Form, forms.Form) and self.permission:
      kw['permission'] = self.permission
    return kw

  def index_url(self):
    return url_for('.index')

  def redirect_to_index(self):
    return redirect(self.index_url())

  @property
  def template_kwargs(self):
    """
    provides :attr:`form` to templates
    """
    kw = super(ObjectView, self).template_kwargs
    kw['form'] = self.form
    return kw


CANCEL_BUTTON = ButtonAction(
  'form', 'cancel', title=_l(u'Cancel'),
  btn_class='default cancel'  # .cancel: if jquery.validate is used it will
)                              # properly skip validation

EDIT_BUTTON = ButtonAction('form', 'edit', btn_class='primary',
                           title=_l(u'Save'))


class ObjectEdit(ObjectView):
  """
  Edit objects
  """
  template = 'default/object_edit.html'
  decorators = (csrf.support_graceful_failure,)
  permission = WRITE

  #: :class:ButtonAction instance to show on form
  _buttons = ()

  #: submitted form data
  data = None

  #: action name from form data
  action = None

  #: button clicked, corresponding to :attr:`action`.
  button = None

  #: verb used to describe activity
  activity_verb = 'update'

  #: UI flash message
  _message_success = _l(u"Entity successfully edited")

  view_endpoint = None

  def __init__(self, Model=None, pk=None, Form=None, template=None,
               view_endpoint=None, message_success=None, *args, **kwargs):
    ObjectView.__init__(self, Model, pk, Form, template=template, *args, **kwargs)
    if view_endpoint is not None:
      self.view_endpoint = view_endpoint

    if not self.view_endpoint:
      self.view_endpoint = '.{}_view'.format(self.Model.__name__)

    if message_success:
      self._message_success = message_success

  def post(self, *args, **kwargs):
    # conservative: no action submitted -> cancel
    action = self.data.get('__action', u'cancel')
    if action == u'cancel':
      return self.cancel()

    return self.handle_action(action)

  def put(self):
    return self.post()

  def prepare_args(self, args, kwargs):
    args, kwargs = super(ObjectEdit, self).prepare_args(args, kwargs)
    self._buttons = self.get_form_buttons(*args, **kwargs)
    self.data = request.form
    return args, kwargs

  def get_form_buttons(self, *args, **kwargs):
    return [EDIT_BUTTON, CANCEL_BUTTON]

  @property
  def buttons(self):
    return (button for button in self._buttons
            if button.available(actions.context))

  def view_url(self):
    kw = { self.pk: self.obj.id }
    return url_for(self.view_endpoint, **kw)

  def redirect_to_view(self):
    if self.button:
      url = self.button.url(actions.context)
      if url:
        return redirect(url)
    return redirect(self.view_url())

  def message_success(self):
    return unicode(self._message_success)

  # actions
  def handle_action(self, action):
    for button in self._buttons:
      if action == button.name:
        if not button.available(dict(view=self)):
          raise ValueError('Action "{}" not available'
                           ''.format(action.encode('utf-8')))
        break
    else:
      raise ValueError('Unknown action: "{}"'.format(action.encode('utf-8')))

    self.action = action
    self.button = button
    return getattr(self, action)()

  def cancel(self):
    return self.redirect_to_view()

  def edit(self):
    if self.validate():
      return self.form_valid()
    else:
      if request.csrf_failed:
        errors = self.form.errors
        csrf_failed = errors.pop('csrf_token', False)
        if csrf_failed and not errors:
          # failed only because of invalid/expired csrf, no error on form
          return self.form_csrf_invalid()

      resp = self.form_invalid()
      if resp:
        return resp

      flash(_(u"Please fix the error(s) below"), "error")

    # if we end here then something wrong has happened: show form with error
    # messages
    return self.get()

  def before_populate_obj(self):
    """
    This method is called after form has been validated and before calling
    `form.populate_obj()`. Sometimes one may want to remove a field from
    the form because it's non-sense to store it on edited object, and use it in
    a specific manner, for example::

        image = form.image
        del form.image
        store_image(image)
    """
    pass

  def after_populate_obj(self):
    """
    Called after `self.obj` values have been updated, and `self.obj`
    attached to an ORM session.
    """
    pass

  def handle_commit_exception(self, exc):
    """
    hook point to handle exception that may happen during commit.

    It is the responsability of this method to perform a rollback if it is
    required for handling `exc`. If the method does not handle `exc` if should
    do nothing and return None.

    :returns: * a valid :class:`Response` if exception is handled.
              * `None` if exception is not handled. Default handling happens.
    """
    return None

  def commit_success(self):
    """
    Called after object has been successfully saved to database
    """

  def validate(self):
    return self.form.validate()

  def form_valid(self):
    """
    Save object.

    Called when form is validated.
    """
    session = current_app.db.session()

    with session.no_autoflush:
      self.before_populate_obj()
      self.form.populate_obj(self.obj)
      session.add(self.obj)
      self.after_populate_obj()

    try:
      session.flush()
      self.send_activity()
      session.commit()
    except ValidationError as e:
      rv = self.handle_commit_exception(e)
      if rv is not None:
        return rv
      session.rollback()
      flash(e.message, "error")
      return self.get()
    except sa.exc.IntegrityError as e:
      rv = self.handle_commit_exception(e)
      if rv is not None:
        return rv
      session.rollback()
      logger.error(e)
      flash(_(u"An entity with this name already exists in the database."),
            "error")
      return self.get()
    else:
      self.commit_success()
      flash(self.message_success(), "success")
      return self.redirect_to_view()

  def form_invalid(self):
    """
    When a form doesn't validate this method is called.

    It may return a :class:`Flask.Response` instance, to handle specific
    errors in custom screens.

    Else the edit form screen is returned with error(s) highlighted.

    This method is useful for detecting edition conflict using hidden fields
    and show a specific screen to help resolve the conflict.
    """
    return None

  def form_csrf_invalid(self):
    """
    Called when a form doesn't validate *only* because of csrf token expiration.

    This works only if form is an instance of :class:`flask_wtf.form.SecureForm`.
    Else default CSRF protection (before request) will take place.

    It must return a valid :class:`Flask.Response` instance. By default it
    returns to edit form screen with an informative message.
    """
    current_app.extensions['csrf-handler'].flash_csrf_failed_message()
    return self.get()

  def send_activity(self):
    activity.send(self,
                  actor=g.user,
                  verb=self.activity_verb,
                  object=self.obj,
                  target=self.activity_target)

  @property
  def activity_target(self):
    """
    Return `target` to use when creating activity.
    """
    return None


CREATE_BUTTON = ButtonAction('form', 'create', btn_class='primary',
                             title=_l(u'Create'))
CHAIN_CREATE_BUTTON = ButtonAction(
    'form', 'chain_create', btn_class='primary',
    title=_l(u'Create and add new'),
    endpoint=lambda ctx: Endpoint(request.endpoint, **request.view_args),
    condition=lambda ctx: getattr(ctx['view'], 'chain_create_allowed', False)
)


class ObjectCreate(ObjectEdit):
  """
  Create a new object
  """
  permission = CREATE
  activity_verb = 'post'
  _message_success = _l(u"Entity successfully added")

  #: set to `True` to show 'Save and add new' button
  chain_create_allowed = False

  def __init__(self, *args, **kwargs):
    chain_create_allowed = kwargs.pop('chain_create_allowed', None)
    if chain_create_allowed is not None:
      self.chain_create_allowed = bool(chain_create_allowed)

    ObjectEdit.__init__(self, *args, **kwargs)

  def prepare_args(self, args, kwargs):
    # we must ensure that no flush() occurs and that obj is not registered in
    # session (to prevent accidental insert of an incomplete object)
    session = current_app.db.session()
    with session.no_autoflush:
      args, kwargs = super(ObjectCreate, self).prepare_args(args, kwargs)

    try:
      session.expunge(self.obj)
    except sa.exc.InvalidRequestError:
      # obj is not in session
      pass

    return args, kwargs

  def init_object(self, args, kwargs):
    self.obj = self.Model()
    return args, kwargs

  def get_form_kwargs(self):
    kw = super(ObjectCreate, self).get_form_kwargs()
    if request.method == 'GET':
      # when GET allow form prefill instead of empty/current object data
      # FIXME: filter allowed parameters on given a field flags (could be
      # 'allow_from_get'?)
      kw['formdata'] = request.args

    return kw

  def get_form_buttons(self, *args, **kwargs):
    return [CREATE_BUTTON, CHAIN_CREATE_BUTTON, CANCEL_BUTTON]

  def breadcrumb(self):
    return nav.BreadcrumbItem(label=CREATE_BUTTON.title)

  # actions
  def create(self):
    return self.edit()

  chain_create = create

  def cancel(self):
    return self.redirect_to_index()


DELETE_BUTTON = ButtonAction('form', 'delete', title=_l(u'Delete'))


class ObjectDelete(ObjectEdit):
  """
  Delete object. Supports DELETE verb.
  """
  methods = ['POST']
  permission = DELETE
  activity_verb = 'delete'
  _message_success = _l(u"Entity deleted")

  init_object = BaseObjectView.init_object

  def get_form_buttons(self, *args, **kwargs):
    return [DELETE_BUTTON, CANCEL_BUTTON]

  def delete(self):
    session = current_app.db.session()
    session.delete(self.obj)
    activity.send(self, actor=g.user, verb="delete", object=self.obj,
                  target=self.activity_target)
    try:
      session.commit()
    except sa.exc.IntegrityError as e:
      rv = self.handle_commit_exception(e)
      if rv is not None:
        return rv
      session.rollback()
      logger.error(e)
      flash(_(u"An entity with this name already exists in the database "
              u"or is referenced by another object and cannot be deleted."),
            "error")
      return self.redirect_to_view()
    else:
      flash(self.message_success(), 'success')
      # FIXME: for DELETE verb response in case of success should be 200, 202
      # (accepted) or 204 (no content)
      return self.redirect_to_index()


class JSONBaseSearch(JSONView):
  Model = None
  minimum_input_length = 2

  def __init__(self, *args, **kwargs):
    Model = kwargs.pop('Model', self.Model)
    minimum_input_length = kwargs.pop('minimum_input_length',
                                      self.minimum_input_length)
    super(JSONBaseSearch, self).__init__(*args, **kwargs)
    self.Model = Model
    self.minimum_input_length = minimum_input_length

  def prepare_args(self, args, kwargs):
    args, kwargs = JSONView.prepare_args(self, args, kwargs)
    kwargs['q'] = kwargs.get("q", u'').replace(u"%", u" ").lower()
    return args, kwargs

  def data(self, q, *args, **kwargs):
    if self.minimum_input_length and len(q) < self.minimum_input_length:
      raise BadRequest(
        'Minimum query length is {:d}'.format(self.minimum_input_length),
      )

    results = []
    for obj in self.get_results(q, **kwargs):
      results.append(self.get_item(obj))

    return dict(results=results)

  def get_results(self, q, *args, **kwargs):
    raise NotImplementedError

  def get_item(self, obj):
    """
    Return a result item

    :param obj: Instance object
    :returns: a dictionnary with at least `id` and `text` values
    """
    raise NotImplementedError


class JSONModelSearch(JSONBaseSearch):
  """
  Base class for json sqlalchemy model search, as used by select2 widgets for
  example
  """

  def get_results(self, q, *args, **kwargs):
    query = self.Model.query

    query = self.options(query)
    query = self.filter(query, q, **kwargs)
    query = self.order_by(query)
    if not q and not self.minimum_input_length:
      query = query.limit(50)
    return query.all()

  def options(self, query):
    return query.options(sa.orm.noload('*'))

  def filter(self, query, q, **kwargs):
    if not q:
      return query
    return query.filter(sa.func.lower(self.Model.name).like(q + "%"))

  def order_by(self, query):
    return query.order_by(self.Model.name)

  def get_label(self, obj):
    return obj.name

  def get_item(self, obj):
    """
    Return a result item

    :param obj: Instance object
    :returns: a dictionnary with at least `id` and `text` values
    """
    return dict(id=obj.id, text=self.get_label(obj), name=obj.name)


class JSONWhooshSearch(JSONBaseSearch):
  """
  Base class for JSON Whoosh search, as used by select2 widgets for example
  """

  def get_results(self, q, *args, **kwargs):
    svc = current_app.services['indexing']
    search_kwargs = {'limit': 30, 'Models': (self.Model,)}
    results = svc.search(q, **search_kwargs)
    return results

  def get_item(self, hit):
    """
    Return a result item

    :param hit: Hit object from Whoosh
    :returns: a dictionnary with at least `id` and `text` values
    """
    return dict(id=hit['id'], text=hit['name'], name=hit['name'])



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import json
import logging
import time
from datetime import timedelta
from uuid import UUID, uuid1

from celery import shared_task
from flask import current_app
from flask_login import AnonymousUserMixin

from abilian.core import signals
from abilian.web import url_for

from .views import bp as blueprint

logger = logging.getLogger(__name__)

CHUNK_SIZE = 64 * 1024

DEFAULT_CONFIG = {
  'USER_QUOTA': 100 * 1024**2, # max 100 Mb for all current files
  'USER_MAX_FILES': 1000, # max number of files per user
  'DELETE_STALLED_AFTER': 60 * 60 * 24 # delete files remaining after 1 day
}

CLEANUP_SCHEDULE_ID = __name__ + '.periodic_clean_upload_directory'
DEFAULT_CLEANUP_SCHEDULE = {
  'task': CLEANUP_SCHEDULE_ID,
  'schedule': timedelta(hours=1),
}


def is_valid_handle(handle):
  try:
    UUID(handle)
  except ValueError:
    return False

  return True


class FileUploadsExtension(object):
  """
  API for Out-Of-Band file uploads.

  Allow to manage files in forms: file is uploaded to an upload url, a handle is
  returned will be used in the form to refer to this uploaded filed.

  If the form fails to validate the uploaded file is not lost.

  A periodic task cleans the temporary repository.
  """
  def __init__(self, app):
    app.extensions['uploads'] = self
    app.add_template_global(self, 'uploads')
    app.register_blueprint(blueprint)
    signals.register_js_api.connect(self._do_register_js_api)

    self.config = {}
    self.config.update(DEFAULT_CONFIG)
    self.config.update(app.config.get('FILE_UPLOADS', {}))
    app.config['FILE_UPLOADS'] = self.config

    # celery schedule
    CELERYBEAT_SCHEDULE = app.config.setdefault('CELERYBEAT_SCHEDULE', {})
    if CLEANUP_SCHEDULE_ID not in CELERYBEAT_SCHEDULE:
      CELERYBEAT_SCHEDULE[CLEANUP_SCHEDULE_ID] = DEFAULT_CLEANUP_SCHEDULE

    path = self.UPLOAD_DIR = app.DATA_DIR / 'uploads'
    if not path.exists():
      path.mkdir(0o775)

    path.resolve()

  def _do_register_js_api(self, sender):
    app = sender
    js_api = app.js_api.setdefault('upload', {})
    js_api['newFileUrl'] = url_for('uploads.new_file')

  def user_dir(self, user):
    user_id = (str(user.id) if not isinstance(user, AnonymousUserMixin)
               else 'anonymous')
    return self.UPLOAD_DIR / user_id

  def add_file(self, user, file_obj, **metadata):
    """
    Add a new file.

    :returns: file handle
    """
    user_dir = self.user_dir(user)
    if not user_dir.exists():
      user_dir.mkdir(0o775)

    handle = str(uuid1())
    file_path = user_dir / handle

    with file_path.open('wb') as out:
      for chunk in iter(lambda: file_obj.read(CHUNK_SIZE), ''):
        out.write(chunk)

    if metadata:
      meta_file = user_dir / '{}.metadata'.format(handle)
      with meta_file.open('wb') as out:
        json.dump(metadata, out, skipkeys=True)

    return handle

  def get_file(self, user, handle):
    """
    Retrieve a file for a user.

    :returns: a :class:`pathlib.Path` instance to this file, or None if no file
    can be found for this handle.
    """
    user_dir = self.user_dir(user)
    if not user_dir.exists():
      return None

    if not is_valid_handle(handle):
      return None

    file_path = user_dir / handle

    if not file_path.exists() and not file_path.is_file():
      return None

    return file_path

  def get_metadata_file(self, user, handle):
    content = self.get_file(user, handle)
    if content is None:
      return None

    metafile = content.parent / '{}.metadata'.format(handle)
    if not metafile.exists():
      return None

    return metafile

  def get_metadata(self, user, handle):
    metafile = self.get_metadata_file(user, handle)
    if metafile is None:
      return {}

    try:
      with metafile.open('rb') as in_:
        meta = json.load(in_)
    except:
      meta = {}

    return meta

  def remove_file(self, user, handle):
    paths = (self.get_file(user, handle),
             self.get_metadata_file(user, handle),)

    for file_path in paths:
      if file_path is not None:
        try:
          file_path.unlink()
        except:
          logger.exception('Error during remove file')

  def clear_stalled_files(self):
    """
    Scan upload directory and delete stalled files.

    Stalled files are files uploaded more than `DELETE_STALLED_AFTER` seconds
    ago.
    """
    # FIXME: put lock in directory?
    CLEAR_AFTER = self.config['DELETE_STALLED_AFTER']
    minimum_age = time.time() - CLEAR_AFTER

    for user_dir in self.UPLOAD_DIR.iterdir():
      if not user_dir.is_dir():
        logger.error('Found non-directory in upload dir: %r', bytes(user_dir))
        continue

      for content in user_dir.iterdir():
        if not content.is_file():
          logger.error('Found non-file in user upload dir: %r', bytes(content))
          continue

        if content.stat().st_ctime < minimum_age:
          content.unlink()

# Task scheduled to run every hour: make it expire after 50min.
@shared_task(expires=3000)
def periodic_clean_upload_directory():
  """
  This task should be run periodically.

  Default config sets up schedule using
  :data:`DEFAULT_CLEANUP_SCHEDULE`. `CELERYBEAT_SCHEDULE` key is
  :data:`CLEANUP_SCHEDULE_ID`.
  """
  uploads = current_app.extensions['uploads']
  uploads.clear_stalled_files()



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from sqlalchemy.types import String, TypeDecorator


class ValueSingletonMeta(type):

  def __new__(cls, name, bases, dct):
    dct['__instances__'] = {}
    dct.setdefault('__slots__', ())
    new_type = type.__new__(cls, name, bases, dct)
    return new_type

  def __call__(cls, value, *args, **kwargs):
    if isinstance(value, cls):
      return value

    if value not in cls.__instances__:
      value_instance = type.__call__(cls, value, *args, **kwargs)
      cls.__instances__[getattr(value_instance, cls.attr)] = value_instance
    return cls.__instances__[value]


class UniqueName(object):
  """
  Base class to create singletons from strings.

  A subclass of :class:`UniqueName` defines a namespace.
  """
  __metaclass__ = ValueSingletonMeta
  __slots__ = ('_hash', '__name')
  attr = 'name'

  def __init__(self, name):
    self.__name = unicode(name).strip().lower()
    self._hash = hash(self.__name)

  @property
  def name(self):
    return self.__name

  def __repr__(self):
    return '{}({})'.format(self.__class__.__name__, repr(self.name))

  def __unicode__(self):
    return self.name

  def __str__(self):
    return self.name.encode(u'utf-8')

  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self._hash == other._hash
    return self.__name == unicode(other)

  def __hash__(self):
    return self._hash


class UniqueNameType(TypeDecorator):
  """
  Sqlalchemy type that stores a subclass of :class:`UniqueName`

  Usage::
    class MySingleton(UniqueName):
        pass

    class MySingletonType(UniqueNameType):
        Type = MySingleton
  """
  impl = String
  Type = None
  default_max_length = 100

  def __init__(self, *args, **kwargs):
    assert self.Type is not None
    kwargs.setdefault('length', self.default_max_length)
    TypeDecorator.__init__(self, *args, **kwargs)

  def process_bind_param(self, value, dialect):
    if value is not None:
      value = str(value)
    return value

  def process_result_value(self, value, dialect):
    if value is not None:
      value = self.Type(value)
    return value



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from flask import current_app, jsonify, send_file
from flask.signals import request_tearing_down
from flask_login import current_user
from flask_wtf.file import FileField, file_required
from werkzeug import secure_filename
from werkzeug.exceptions import BadRequest, NotFound

from abilian.core.util import pdb_on_error
from abilian.web import csrf, url_for
from abilian.web.blueprints import Blueprint
from abilian.web.forms import Form
from abilian.web.views import JSONView, View

bp = Blueprint('uploads', __name__, url_prefix='/upload')

class UploadForm(Form):

  file = FileField(validators=(file_required(),))


class BaseUploadsView(object):

  def prepare_args(self, args, kwargs):
    args, kwargs = super(BaseUploadsView, self).prepare_args(args, kwargs)
    self.uploads = current_app.extensions['uploads']
    self.user = current_user._get_current_object()
    return args, kwargs


class NewUploadView(BaseUploadsView, JSONView):
  """
  Upload a new file
  """
  methods = ['POST', 'PUT']
  decorators = (csrf.support_graceful_failure,)

  #: file handle to be returned
  handle = None

  def data(self, *args, **kwargs):
    return {
      'handle': self.handle,
      'url': url_for('.handle', handle=self.handle),
    }

  def post(self, *args, **kwargs):
    form = UploadForm()

    if not form.validate():
      raise BadRequest('File is missing.')

    uploaded = form['file'].data
    filename = secure_filename(uploaded.filename)
    mimetype = uploaded.mimetype
    self.handle = self.uploads.add_file(self.user, uploaded,
                                        filename=filename,
                                        mimetype=mimetype)
    return self.get(*args, **kwargs)

  def put(self, *args, **kwargs):
    return self.post(*args, **kwargs)


bp.add_url_rule('/', view_func=NewUploadView.as_view('new_file',),)

class UploadView(BaseUploadsView, View):
  """
  Manage an uploaded file: download, delete
  """
  methods = ['GET', 'DELETE']
  decorators = (csrf.support_graceful_failure,)

  def get(self, handle, *args, **kwargs):
    file_obj = self.uploads.get_file(self.user, handle)

    if file_obj is None:
      raise NotFound()

    metadata = self.uploads.get_metadata(self.user, handle)
    filename = metadata.get('filename', handle)
    content_type = metadata.get('mimetype')
    stream = file_obj.open('rb')

    return send_file(stream,
                     as_attachment=True,
                     attachment_filename=filename,
                     mimetype=content_type,
                     cache_timeout=0,
                     add_etags=False,
                     conditional=False)

  def delete(self, handle, *args, **kwargs):
    if self.uploads.get_file(self.user, handle) is None:
      raise NotFound()

    self.uploads.remove_file(self.user, handle)
    return jsonify({'success': True})


bp.add_url_rule('/<string:handle>', view_func=UploadView.as_view('handle',),)



# coding=utf-8
"""
"""
from __future__ import absolute_import, print_function, division

from flask import render_template

from abilian.i18n import _, _l
from abilian.web.util import url_for
from abilian.web.admin.panel import AdminPanel

from . import views


class UsersPanel(AdminPanel):
  """
  User administration panel.
  """
  id = 'users'
  label = _l(u'Users')
  icon = 'user'

  def install_additional_rules(self, add_url_rule):
    add_url_rule('/users', view_func=views.JsonUsersList.as_view('json_list'))
    add_url_rule('/new',
                 view_func=views.UserCreate.as_view('new'))
    add_url_rule('/<int:user_id>',
                 view_func=views.UserEdit.as_view('user'))

  def get(self):
    # FIXME: use widgets.AjaxMainTableView instead
    datatable_options = {
      'sDom': 'lfFritip',
      'aaSorting': [
        [1, u'asc'],
      ],
      'aoColumns': [
        dict(bSortable=False),
        dict(asSorting=['asc', 'desc']),
        dict(asSorting=['asc', 'desc']),
        dict(bSortable=False),
        dict(bSortable=False),
        dict(bSortable=False),
        dict(asSorting=['asc', 'desc']),
      ],
      'bFilter': True,
      'oLanguage': {
        'sSearch': _("Filter records:"),
        'sPrevious': _("Previous"),
        'sNext': _("Next"),
        'sInfo': _("Showing _START_ to _END_ of _TOTAL_ entries"),
        'sInfoFiltered': _("(filtered from _MAX_ total entries)"),
        'sAddAdvancedFilter': _("Add a filter"),
      },
      'bStateSave': False,
      'bPaginate': True,
      'sPaginationType': "bootstrap",
      'bLengthChange': False,
      'iDisplayLength': 30,
      'bProcessing': True,
      'bServerSide': True,
      'sAjaxSource': url_for('.users_json_list'),
    }

    return render_template('admin/users.html',
                           next=next,
                           datatable_options=datatable_options)



# coding=utf-8
""" Additional data types for sqlalchemy
"""
from __future__ import absolute_import, division, print_function

import json
import logging
import sys
import uuid
from distutils.version import StrictVersion
from functools import partial

import babel
import flask_sqlalchemy as flask_sa
import pkg_resources
import pytz
import sqlalchemy as sa
from flask_sqlalchemy import SQLAlchemy as SAExtension
from future.utils import string_types
from sqlalchemy.ext.mutable import Mutable

from .logging import patch_logger

logger = logging.getLogger(__name__)

FLASK_SA_VERSION = pkg_resources.get_distribution('Flask-SQLAlchemy').version


@sa.event.listens_for(sa.pool.Pool, "checkout")
def ping_connection(dbapi_connection, connection_record, connection_proxy):
  """
  Ensure connections are valid.

  From: `http://docs.sqlalchemy.org/en/rel_0_8/core/pooling.html`

  In case db has been restarted pool may return invalid connections.
  """
  cursor = dbapi_connection.cursor()
  try:
    cursor.execute("SELECT 1")
  except:
    # optional - dispose the whole pool
    # instead of invalidating one at a time
    # connection_proxy._pool.dispose()

    # raise DisconnectionError - pool will try
    # connecting again up to three times before raising.
    raise sa.exc.DisconnectionError()
  cursor.close()


class AbilianBaseSAExtension(SAExtension):
  """
  Base subclass of :class:`flask_sqlalchemy.SQLAlchemy`. Add
  our custom driver hacks.
  """
  def apply_driver_hacks(self, app, info, options):
    SAExtension.apply_driver_hacks(self, app, info, options)

    if info.drivername == 'sqlite':
      connect_args = options.setdefault('connect_args', {})

      if 'isolation_level' not in connect_args:
        # required to support savepoints/rollback without error. It disables
        # implicit BEGIN/COMMIT statements made by pysqlite (a COMMIT kills all
        # savepoints made).
        connect_args['isolation_level'] = None
    elif info.drivername.startswith('postgres'):
      options.setdefault('client_encoding', 'utf8')


if StrictVersion(FLASK_SA_VERSION) <= StrictVersion('1.0'):
  # SA extension's scoped session supports 'bind' parameter only after 1.0. This
  # is a fix for it. This is required to ensure transaction rollback during
  # tests, but it's useful in some use cases too.
  from flask_sqlalchemy import _SignallingSession as BaseSession

  class SignallingSession(BaseSession):

    def __init__(self, db, autocommit=False, autoflush=True, **options):
      self.app = db.get_app()
      self._model_changes = {}
      bind = options.pop('bind', None) or db.engine
      # actually we are overriding BaseSession.__init__, so we don't want to
      # call it! Directly call BaseSession parent __init__
      sa.orm.Session.__init__(self, autocommit=autocommit, autoflush=autoflush,
                              bind=bind, binds=db.get_binds(self.app),
                              **options)

  class SQLAlchemy(AbilianBaseSAExtension):

    def create_scoped_session(self, options=None):
      """Helper factory method that creates a scoped session."""
      # override needed to use our SignallingSession implementation
      if options is None:
        options = {}
      scopefunc = options.pop('scopefunc', None)
      return sa.orm.scoped_session(partial(SignallingSession, self, **options),
                                   scopefunc=scopefunc)


else:
  # Flask-SQLAlchemy > 1.0: bind parameter is supported
  SQLAlchemy = AbilianBaseSAExtension

del FLASK_SA_VERSION


# PATCH flask_sqlalchemy for proper info in debug toolbar.
#
# Original code works only when current app code is involved. If using 3rd party
# app the query is logged but source is marked "unknown". Our patch is a "best
# guess".
def _calling_context(app_path):
  frm = sys._getframe(1)
  entered_sa_code = exited_sa_code = False
  sa_caller = '<unknown>'
  format_name = ('{frm.f_code.co_filename}:{frm.f_lineno} '
                 '({frm.f_code.co_name})'.format)

  while frm.f_back is not None:
    name = frm.f_globals.get('__name__')
    if name and (name == app_path
                 or name.startswith(app_path + '.')
                 or name.startswith('abilian.')):
      return format_name(frm=frm)

    if not exited_sa_code:
      in_sa_code = name and (name == 'sqlalchemy'
                             or name.startswith('sqlalchemy.'))
      if not entered_sa_code:
        entered_sa_code = in_sa_code
      elif not in_sa_code:
        # exited from sa stack: retain name
        sa_caller = format_name(frm=frm)
        exited_sa_code = True

    frm = frm.f_back

  return sa_caller

patch_logger.info(flask_sa._calling_context)
flask_sa._calling_context = _calling_context
del flask_sa
# END PATCH


def filter_cols(model, *filtered_columns):
  """
  Return columnsnames for a model except named ones. Useful for defer()
  for example to retain only columns of interest
  """
  m = sa.orm.class_mapper(model)
  return list(set(p.key for p in m.iterate_properties
              if hasattr(p, 'columns')).difference(filtered_columns))


class MutationDict(Mutable, dict):
  """Provides a dictionary type with mutability support."""

  @classmethod
  def coerce(cls, key, value):
    """Convert plain dictionaries to MutationDict."""
    if not isinstance(value, MutationDict):
      if isinstance(value, dict):
        return MutationDict(value)

      # this call will raise ValueError
      return Mutable.coerce(key, value)
    else:
      return value

  #  pickling support. see:
  #  http://docs.sqlalchemy.org/en/rel_0_8/orm/extensions/mutable.html#supporting-pickling
  def __getstate__(self):
    return dict(self)

  def __setstate__(self, state):
    self.update(state)

  # dict methods
  def __setitem__(self, key, value):
    """Detect dictionary set events and emit change events."""
    dict.__setitem__(self, key, value)
    self.changed()

  def __delitem__(self, key):
    """Detect dictionary del events and emit change events."""
    dict.__delitem__(self, key)
    self.changed()

  def clear(self):
    dict.clear(self)
    self.changed()

  def update(self, other):
    dict.update(self, other)
    self.changed()

  def setdefault(self, key, failobj=None):
    if key not in self:
      self.changed()
    return dict.setdefault(self, key, failobj)

  def pop(self, key, *args):
    self.changed()
    return dict.pop(self, key, *args)

  def popitem(self):
    self.changed()
    return dict.popitem(self)


class MutationList(Mutable, list):
  """
  Provides a list type with mutability support.
  """
  @classmethod
  def coerce(cls, key, value):
    """Convert list to MutationList."""
    if not isinstance(value, MutationList):
      if isinstance(value, list):
        return MutationList(value)

      # this call will raise ValueError
      return Mutable.coerce(key, value)
    else:
      return value

  #  pickling support. see:
  #  http://docs.sqlalchemy.org/en/rel_0_8/orm/extensions/mutable.html#supporting-pickling
  def __getstate__(self):
    d = self.__dict__.copy()
    d.pop('_parents', None)
    return d

  # list methods
  def __setitem__(self, idx, value):
    list.__setitem__(self, idx, value)
    self.changed()

  def __delitem__(self, idx):
    list.__delitem__(self, idx)
    self.changed()

  def insert(self, idx, value):
    list.insert(self, idx, value)
    self.changed()

  def __setslice__(self, i, j, other):
    list.__setslice__(self, i, j, other)
    self.changed()

  def __delslice__(self, i, j):
    list.__delslice__(self, i, j)
    self.changed()

  def __iadd__(self, other):
    l = list.__iadd__(self, other)
    self.changed()
    return l

  def __imul__(self, n):
    l = list.__imul__(self, n)
    self.changed()
    return l

  def append(self, item):
    list.append(self, item)
    self.changed()

  def pop(self, i=-1):
    item = list.pop(self, i)
    self.changed()
    return item

  def remove(self, item):
    list.remove(self, item)
    self.changed()

  def reverse(self):
    list.reverse(self)
    self.changed()

  def sort(self, *args, **kwargs):
    list.sort(self, *args, **kwargs)
    self.changed()

  def extend(self, other):
    list.extend(self, other)
    self.changed()


class JSON(sa.types.TypeDecorator):
  """Stores any structure serializable with json.

  Usage
    JSON()
    Takes same parameters as sqlalchemy.types.Text
  """
  impl = sa.types.Text

  def process_bind_param(self, value, dialect):
    if value is not None:
      value = json.dumps(value)
    return value

  def process_result_value(self, value, dialect):
    if value is not None:
      value = json.loads(value)
    return value


class JSONUniqueListType(JSON):
  """ Store a list in JSON format, with items made unique and sorted.
  """
  @property
  def python_type(self):
    return MutationList

  def process_bind_param(self, value, dialect):
    # value may be a simple string used in a LIKE clause for instance, so we
    # must ensure we uniquify/sort only for list-like values
    if value is not None and isinstance(value, (tuple, list)):
      value = sorted(set(value))

    return JSON.process_bind_param(self, value, dialect)


def JSONDict(*args, **kwargs):
  """
  Stores a dict as JSON on database, with mutability support.
  """
  return MutationDict.as_mutable(JSON(*args, **kwargs))


def JSONList(*args, **kwargs):
  """
  Stores a list as JSON on database, with mutability support.

  If kwargs has a param `unique_sorted` (which evaluated to True), list values
  are made unique and sorted.
  """
  type_ = JSON
  try:
    if kwargs.pop('unique_sorted'):
      type_ = JSONUniqueListType
  except KeyError:
    pass

  return MutationList.as_mutable(type_(*args, **kwargs))


class UUID(sa.types.TypeDecorator):
  """
  Platform-independent UUID type.

  Uses Postgresql's UUID type, otherwise uses
  CHAR(32), storing as stringified hex values.

  From SQLAlchemy documentation.
  """
  impl = sa.types.CHAR

  def load_dialect_impl(self, dialect):
    if dialect.name == 'postgresql':
      return dialect.type_descriptor(sa.dialects.postgresql.UUID())
    else:
      return dialect.type_descriptor(sa.types.CHAR(32))

  def process_bind_param(self, value, dialect):
    if value is None:
      return value
    elif dialect.name == 'postgresql':
      return str(value)
    else:
      if not isinstance(value, uuid.UUID):
        value = uuid.UUID(value)
      # hexstring
      return "%.32x" % value

  def process_result_value(self, value, dialect):
    return value if value is None else uuid.UUID(value)

  def compare_against_backend(self, dialect, conn_type):
    if dialect.name == 'postgresql':
      return isinstance(conn_type, sa.dialects.postgresql.UUID)
    else:
      return isinstance(conn_type, sa.types.CHAR)


class Locale(sa.types.TypeDecorator):
  """
  Store a :class:`babel.Locale` instance
  """
  impl = sa.types.UnicodeText

  @property
  def python_type(self):
    return babel.Locale

  def process_bind_param(self, value, dialect):
    if value is None:
      return None

    if not isinstance(value, babel.Locale):
      if not isinstance(value, string_types):
        raise ValueError("Unknown locale value: %s" % repr(value))
      if not value.strip():
        return None
      value = babel.Locale.parse(value)

    code = unicode(value.language)
    if value.territory:
      code += u'_' + unicode(value.territory)
    elif value.script:
      code += u'_' + unicode(value.territory)

    return code

  def process_result_value(self, value, dialect):
    return None if value is None else babel.Locale.parse(value)


class Timezone(sa.types.TypeDecorator):
  """
  Store a :class:`pytz.tzfile.DstTzInfo` instance
  """
  impl = sa.types.UnicodeText

  @property
  def python_type(self):
    return pytz.tzfile.DstTzInfo

  def process_bind_param(self, value, dialect):
    if value is None:
      return None

    if not isinstance(value, pytz.tzfile.DstTzInfo):
      if not isinstance(value, string_types):
        raise ValueError("Unknown timezone value: %s" % repr(value))
      if not value.strip():
        return None
      value = babel.dates.get_timezone(value)

    return value.zone

  def process_result_value(self, value, dialect):
    return None if value is None else babel.dates.get_timezone(value)


# SQLAlchemy > 0.9 has a function generator that forget to set __module__
# attributes in 0.9.x series. This is fixed in 1.0.x. Sphinx will try to include
# those symbols in documentation, and this may break since we don't have
# sphinx's extensions used by sqlalchemy author.
#
# Ref: https://bitbucket.org/zzzeek/sqlalchemy/issues/3218/__module__-should-be-set-on-functions
if not sa.orm.relationship.__module__:
  sa.orm.relationship.__module__ = 'sqlalchemy.orm'



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from wtforms.fields import StringField

from abilian.core.models.attachment import Attachment
from abilian.i18n import _l
from abilian.web.forms import Form
from abilian.web.forms.fields import FileField
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import required


class AttachmentForm(Form):

  blob = FileField(
    _l(u'file'),
    validators=[required()],
    filters=(strip,),
    multiple=False)

  description = StringField(
    _l(u'description (optional)'),
    filters=(strip,),
  )

  class Meta:
    model = Attachment
    include_primary_keys = True
    assign_required = False # for 'id': allow None, for new records



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

import abc

import sqlalchemy as sa
from sqlalchemy import Column, ForeignKey, Integer, UnicodeText
from sqlalchemy.orm import backref, relationship

from abilian.core.entities import Entity

from .blob import Blob

#: name of backref on target :class:`Entity` object
ATTRIBUTE = '__attachments__'


class SupportAttachment(object):
  """
  """
  __metaclass__ = abc.ABCMeta


def register(cls):
  """
  Register an :class:`~.Entity` as a attachmentable class.

  Can be used as a class decorator:

  .. code-block:: python

      @attachment.register
      class MyContent(Entity):
          ....
  """
  if not issubclass(cls, Entity):
    raise ValueError('Class must be a subclass of abilian.core.entities.Entity')

  SupportAttachment.register(cls)
  return cls


def is_support_attachments(obj):
  """
  :param obj: a class or instance
  """
  if isinstance(obj, type):
    return issubclass(obj, SupportAttachment)

  if not isinstance(obj, SupportAttachment):
    return False

  if obj.id is None:
    return False

  return True


def for_entity(obj, check_support_attachments=False):
  """
  Return attachments on an entity.
  """
  if check_support_attachments and not is_support_attachments(obj):
    return []

  return getattr(obj, ATTRIBUTE)


class Attachment(Entity):
  """
  An Attachment owned by an :class:`Entity`.
  """
  __auditable_entity__ = ('entity',
                          'attachment',
                          ('id', 'name'))

  @sa.ext.declarative.declared_attr
  def __mapper_args__(cls):
    # we cannot use super(Attachment, cls): declared_attr happens during class
    # construction. super(cls, cls) could work; as long as `cls` is not a
    # subclass of `Attachment`: it would enter into an infinite loop.
    #
    # Entity.__mapper_args__ calls the descriptor with 'Entity', not `cls`.
    args = Entity.__dict__['__mapper_args__'].fget(cls)
    args['order_by'] = cls.created_at
    return args

  entity_id = Column(Integer, ForeignKey(Entity.id), nullable=False)

  #: owning entity
  entity = relationship(
    Entity,
    lazy='immediate',
    foreign_keys=[entity_id],
    backref=backref(ATTRIBUTE,
                    lazy='select',
                    order_by='Attachment.created_at',
                    cascade="all, delete-orphan",
                  )
  )

  blob_id = Column(Integer, sa.ForeignKey(Blob.id), nullable=False)
  #: file. Stored in a :class:`Blob`
  blob = relationship(Blob, cascade='all, delete', foreign_keys=[blob_id])

  description = Column(UnicodeText(), nullable=False,
                       default=u'', server_default=u'',)

  def __repr__(self):
    class_ = self.__class__
    mod_ = class_.__module__
    classname = class_.__name__
    return '<{}.{} instance at 0x{:x} entity id={!r}'\
        .format(mod_, classname, id(self), self.entity_id)


@sa.event.listens_for(Attachment, 'before_insert', propagate=True)
@sa.event.listens_for(Attachment, 'before_update', propagate=True)
def set_attachment_name(mapper, connection, target):
  if target.name:
    return

  blob = target.blob
  if not blob:
    return

  filename = blob.meta.get('filename')
  if filename is not None:
    target.name = filename



from __future__ import absolute_import, division, print_function

import datetime

import html5lib
import mock
from flask import Flask
from flask_babel import Babel
from flask_testing import TestCase as FlaskTestCase
from jinja2 import Environment
from pytz import timezone, utc

from .. import filters

env = Environment()
filters.init_filters(env)


def en_locale():
  return 'en'


def user_tz():
  # This one is GMT+8 and has no DST (tests should pass any time in year)
  return 'Asia/Hong_Kong'

USER_TZ = timezone(user_tz())


class TestFilters(FlaskTestCase):

  def create_app(self):
    app = Flask(__name__)
    babel = Babel(app, default_locale='fr', default_timezone=USER_TZ)
    babel.localeselector(en_locale)
    babel.timezoneselector(user_tz)
    return app

  def test_labelize(self):
    labelize = filters.labelize
    self.assertEquals(u'Test Case', labelize(u'test_case'))

  def test_filesize(self):
    filesize = filters.filesize
    self.assertEquals("100&nbsp;B", str(filesize('100')))
    self.assertEquals("100&nbsp;B", str(filesize(100)))
    self.assertEquals("1.0&nbsp;kB", str(filesize(1000)))
    self.assertEquals("1.1&nbsp;kB", str(filesize(1100)))
    self.assertEquals("10&nbsp;kB", str(filesize(10000)))
    self.assertEquals("1.1&nbsp;MB", str(filesize(1100100)))
    self.assertEquals("10&nbsp;MB", str(filesize(10000000)))
    self.assertEquals("1.1&nbsp;GB", str(filesize(1100100000)))
    self.assertEquals("100&nbsp;GB", str(filesize(100000000000)))

  def test_roughsize(self):
    roughsize = filters.roughsize
    self.assertEquals(u'6', roughsize(6))
    self.assertEquals(u'15', roughsize(15))
    self.assertEquals(u'130+', roughsize(134))
    self.assertEquals(u'10+', roughsize(15, above=10))
    self.assertEquals(u'55+', roughsize(57, mod=5))

  def test_date_age(self):
    date_age = filters.date_age
    now = datetime.datetime(2012, 6, 10, 10, 10, 10, tzinfo=utc)

    self.assertEquals("", date_age(None))
    dt = datetime.datetime(2012, 6, 10, 10, 10, 0, tzinfo=utc)
    self.assertEquals("2012-06-10 18:10 (1 minute ago)", date_age(dt, now))

    dt = datetime.datetime(2012, 6, 10, 10, 8, 10, tzinfo=utc)
    self.assertEquals("2012-06-10 18:08 (2 minutes ago)", date_age(dt, now))

    dt = datetime.datetime(2012, 6, 10, 8, 30, 10, tzinfo=utc)
    self.assertEquals("2012-06-10 16:30 (2 hours ago)", date_age(dt, now))

    # for coverage: test when using default parameter now=None
    dt_patcher = mock.patch.object(filters.datetime, 'datetime',
                                   mock.Mock(wraps=datetime.datetime))
    with dt_patcher as mocked:
      mocked.utcnow.return_value = now
      self.assertEquals("2012-06-10 16:30 (2 hours ago)", date_age(dt))

  def test_age(self):
    age = filters.age
    now = datetime.datetime(2012, 6, 10, 10, 10, 10, tzinfo=utc)
    d1m =  datetime.datetime(2012, 6, 10, 10, 10, 0, tzinfo=utc)
    d3w = datetime.datetime(2012, 5, 18, 8, 0, 0, tzinfo=utc)
    d2011 = datetime.datetime(2011, 9, 4, 12, 12, 0, tzinfo=utc)

    # default parameters
    assert age(None) == u''
    assert age(d1m, now) == u'1 minute ago'
    assert age(d3w, now) == u'3 weeks ago'

    # with direction
    assert age(d1m, now, add_direction=False) == u'1 minute'
    assert age(d3w, now, add_direction=False) == u'3 weeks'

    # with date_threshold
    assert age(d1m, now, date_threshold='day') == u'1 minute ago'
    # same year: 2012 not shown
    assert age(d3w, now, date_threshold='day') == u'May 18, 4:00 PM'
    # different year: 2011 shown
    assert (age(d2011, now, date_threshold='day') ==
            u'September 4, 2011, 8:12 PM')

    # using default parameter now=None
    dt_patcher = mock.patch.object(filters.datetime, 'datetime',
                                   mock.Mock(wraps=datetime.datetime))
    with dt_patcher as mocked:
      mocked.utcnow.return_value = now
      assert age(d1m) == u'1 minute ago'


  def test_abbrev(self):
    abbrev = filters.abbrev
    self.assertEquals(u'test', abbrev(u'test', 20))
    self.assertEquals(u'Longer test...e truncated',
                      abbrev(u'Longer test. it should be truncated', 25))

  def test_linkify(self):
    tmpl = env.from_string('{{ "http://test.example.com"|linkify}}')
    rendered = tmpl.render()
    el = html5lib.parseFragment(rendered)
    self.assertEquals(len(el.getchildren()), 1)
    el = el.getchildren()[0]
    self.assertEquals(el.tag, u'{http://www.w3.org/1999/xhtml}a')
    self.assertEquals(el.text, u'http://test.example.com')
    self.assertEquals(
      sorted(el.items()),
      [(u'href', u'http://test.example.com'), (u'rel', u'nofollow')]
    )

  def test_nl2br(self):
    tmpl = env.from_string(
      '{{ "first line\nsecond line\n\n  third, indented" | nl2br }}')
    self.assertEquals(tmpl.render(),
        u'first line<br />\nsecond line<br />\n<br />\n  third, indented')

  def test_paragraphs(self):
    tmpl = env.from_string(
    '''{{ "First paragraph
    some text
    with line return

    Second paragraph
    ... lorem

    Last one - a single line" | paragraphs }}
    ''')

    self.assertEquals(tmpl.render(),
        u'''<p>First paragraph<br />
    some text<br />
    with line return</p>

<p>Second paragraph<br />
    ... lorem</p>

<p>Last one - a single line</p>
    ''')



# coding=utf-8
"""
"""
from __future__ import absolute_import, division, print_function

from abilian.core.models.tag import TAGS_ATTR, Tag, is_support_tagging
from abilian.i18n import _l
from abilian.web import url_for
from abilian.web.forms import Form
from abilian.web.views.object import EDIT_BUTTON

from .forms import TagsField
from .views import bp as tags_bp
from .views import entity_bp

ENTITY_DEFAULT_NS_ATTR = '__tags_default_ns__'

def ns(ns):
  """
  Class decorator that set default tags namespace to use with its instances.
  """
  def setup_ns(cls):
    setattr(cls, ENTITY_DEFAULT_NS_ATTR, ns)
    return cls

  return setup_ns


class _TagsForm(Form):
  """
  Form to workaround a wtforms limitation: fields cannot start with an
  underscore, so '__tags__' is not accepted.

  This form help process tags (and only tags).

  .. seealso:: :py:meth:`~TagsExtension.entity_tags_form`
  """
  def process(self, formdata=None, obj=None, data=None, **kwargs):
    tags = None
    if obj is not None:
      tags = getattr(obj, TAGS_ATTR, set())

    super(_TagsForm, self).process(formdata=formdata, obj=None, data=data,
                                   tags=tags, **kwargs)


class TagsExtension(object):
  """
  API for tags, installed as an application extension

  It is also available in templates as `tags`
  """
  def __init__(self, app):
    app.extensions['tags'] = self
    app.add_template_global(self, 'tags')
    app.register_blueprint(tags_bp)
    app.register_blueprint(entity_bp)

  def is_support_tagging(self, entity):
    return is_support_tagging(entity)

  def entity_tags(self, entity):
    return getattr(entity, TAGS_ATTR)

  def tags_from_hit(self, tag_ids):
    """
    :param tag_ids: indexed ids of tags in hit result. Do not pass hit instance.

    :returns: an iterable of :class:`Tag` instances.
    """
    ids = []
    for t in tag_ids.split():
      t = t.strip()
      try:
        t = int(t)
      except ValueError:
        pass
      else:
        ids.append(t)

    if not ids:
      return []

    return Tag.query.filter(Tag.id.in_(ids)).all()

  def entity_default_ns(self, entity):
    return getattr(entity, ENTITY_DEFAULT_NS_ATTR, 'default')

  def entity_tags_form(self, entity, ns=None):
    """
    Construct a form class with a field for tags in namespace `ns`
    """
    if ns is None:
      ns = self.entity_default_ns(entity)

    field  = TagsField(label=_l(u'Tags'), ns=ns)
    cls = type('EntityNSTagsForm', (_TagsForm,), {'tags': field})
    return cls

  def get(self, ns, label=None):
    """
    Return :class:`tags instances<~Tag>` for the namespace `ns`, ordered by
    label.

    If `label` is not None the only one instance may be returned, or `None` if
    no tags exists for this label.
    """
    query = Tag.query.filter(Tag.ns == ns)

    if label is not None:
      return query.filter(Tag.label == label).first()

    return query.all()

  def add(self, entity, tag=None, ns=None, label=None):
    """
    :return: tag
    """
    if tag is None:
      assert None not in (ns, label)
      tag = self.get(ns, label)
      if tag is None:
        tag = Tag(ns=ns, label=label)

    tags = self.entity_tags(entity)
    tags.add(tag)
    return tag

  def remove(self, entity, tag=None, ns=None, label=None):
    if tag is None:
      assert None not in (ns, label)
      tag = self.get(ns, label)

    tags = self.entity_tags(entity)
    try:
      tags.remove(tag)
    except KeyError:
      pass

  def get_form_context(self, obj, ns=None):
    """
    Return a dict: form instance, action button, submit url...
    Used by macro m_tags_form(entity)
    """
    ctx = {}
    ctx['url'] = url_for('entity_tags.edit', object_id=obj.id)
    ctx['form'] = self.entity_tags_form(obj)(obj=obj, ns=ns)
    ctx['buttons'] = [EDIT_BUTTON]
    return ctx



# coding=utf-8
"""
I18n.

To mark strings for transalation::

    from abilian.i18n import _
    _(u'message to translate')

Use :data:`_` for gettext, :data:`_l` for lazy_gettext, :data:`_n` for
ngettext.

:data:`Babel extension<babel>` support multiple translation paths. This
allows to add more catalogs to search for translations, in LIFO
order. This feature can be used to override some translations in a
custom application, be providing a catalog with messages to override::

    current_app.extensions['babel'].add_translations('abilian.core')

See :meth:`add_translations<Babel.add_translations>`.

To extract messages to build the message catalog template (.pot), use the
following "`-k`" parameters:

.. code-block:: bash

    $ pybabel extract -F babel.cfg -k "_n:1,2" -k "_l" -o "msg.pot" "src"


This can be made easier by placing in `setup.cfg`:

.. code-block:: ini

    [extract_messages]
    mapping_file = babel.cfg
    keywords = _n:1,2 _l
    output-file = msg.pot
    input-dirs = src


And just type:

.. code-block:: bash

    $ python setup.py extract_messages


"""
from __future__ import absolute_import, division, print_function

import importlib
import os
import re
import unicodedata
from contextlib import contextmanager
from datetime import datetime
from gettext import GNUTranslations

import flask_babel
import pytz
from babel import Locale
from babel.dates import LOCALTZ, get_timezone, get_timezone_gmt
from babel.localedata import locale_identifiers
from babel.support import Translations as BaseTranslations
from flask import _request_ctx_stack, current_app, g, render_template, request
from flask_babel import Babel as BabelBase
from flask_babel import gettext, lazy_gettext, ngettext
from future.utils import string_types
from pathlib import Path

__all__ = [
  'babel',
  'gettext', '_',
  'lazy_gettext', '_l', 'localeselector',
  'ngettext', '_n',
  'set_locale',
  'timezoneselector',
  'VALID_LANGUAGES_CODE',
  'render_template_i18n'
]

#: gettext alias
_ = gettext

#: lazy_gettext alias
_l = lazy_gettext
#: ngettext alias
_n = ngettext

#: accepted languages codes
VALID_LANGUAGES_CODE = frozenset(lang for lang in locale_identifiers()
                                 if len(lang) == 2)

def get_default_locale():
  return current_app.extensions['babel'].default_locale


def _get_locale():
  locale = flask_babel.get_locale()
  if locale is None:
    locale = get_default_locale()
  return locale


def __gettext_territory(code):
  locale = _get_locale()
  return (
      locale.territories.get(code)
      or get_default_locale().territories.get(code))

#: get localized territory name
def country_name(code):
  return __gettext_territory(code)

#: lazy version of :func:`country_name`
def lazy_country_name(code):
  from speaklater import make_lazy_string
  return make_lazy_string(__gettext_territory, code)


def default_country():
  return current_app.config.get('DEFAULT_COUNTRY')


def country_choices(first=None, default_country_first=True):
  """
  Return a list of (code, countries), alphabetically sorted on localized
  country name.

  :param first: Country code to be placed at the top
  """
  locale = _get_locale()
  territories = [(code, name) for code, name in locale.territories.iteritems()
                 if len(code) == 2] # skip 3-digit regions

  if first is None and default_country_first:
    first = default_country()

  def sortkey(item):
    if first is not None and item[0] == first:
      return u'0'
    return to_lower_ascii(item[1])

  territories.sort(key=sortkey)
  return territories


def supported_app_locales():
  """
  language codes and labels supported by current application

  :return: an iterable of `(:class:`babel.Locale`, label)`, label being the
  locale language human name in current locale.
  """
  locale = _get_locale()
  codes = current_app.config['BABEL_ACCEPT_LANGUAGES']
  return ((Locale.parse(code), locale.languages.get(code, code))
          for code in codes)


def timezones_choices():
  """
  Timezones values and their labels for current locale.

  :return: an iterable of `(code, label)`, code being a timezone code and label
  the timezone name in current locale.
  """
  utcnow = pytz.utc.localize(datetime.utcnow())
  locale = _get_locale()
  for tz in sorted(pytz.common_timezones):
    tz = get_timezone(tz)
    now = tz.normalize(utcnow.astimezone(tz))
    label = u'({}) {}'.format(get_timezone_gmt(now, locale=locale), tz.zone)
    yield (tz, label) #get_timezone_name(tz, locale=locale))


class Babel(BabelBase):
  """
  Allow to load translations from other modules
  """
  _translations_paths = None

  def __init__(self, *args, **kwargs):
    BabelBase.__init__(self, *args, **kwargs)

  def init_app(self, app):
    super(Babel, self).init_app(app)
    self._translations_paths = [
      (os.path.join(app.root_path, 'translations'), 'messages')
    ]

  def add_translations(self, module_name,
                       translations_dir='translations',
                       domain='messages'):
    """
    Adds translations from external module. For example::

        babel.add_translations('abilian.core')

    Will add translations files from `abilian.core` module.
    """
    module = importlib.import_module(module_name)
    for path in (Path(p, translations_dir) for p in module.__path__):
      if not (path.exists() and path.is_dir()):
        continue

      if not os.access(unicode(path), os.R_OK):
        self.app.logger.warning(
          "Babel translations: read access not allowed {}, skipping."
          "".format(
            repr(unicode((path)).encode('utf-8')))
        )
        continue

      self._translations_paths.append((unicode(path), domain))


class Translations(BaseTranslations):
  """
  Merge only non-empty translations.

  This avoids having uncomplete catalog that "clear" existing translations, when
  used with :func:`_get_translations_multi_paths`.
  """
  def merge(self, translations):
    if isinstance(translations, GNUTranslations):

      for msgkey, msgstr in translations._catalog.items():
        msgid = msgkey

        if isinstance(msgkey, tuple):
          msgid = msgkey[0]

        msgstr = msgstr.strip()
        if msgkey in self._catalog and (msgid == msgstr):
          # when msgstr is empty, compile_catalog sets msgstr = msgid so this is
          # probable an existing translation that would be "erased" by msgid
          # string: skip it.

          # logger.debug('Catalog: %r, skip msgkey: %r, existing: %r',
          #                translations, msgkey, self._catalog[msgkey])
          continue

        self._catalog[msgkey] = msgstr

      if isinstance(translations, BaseTranslations):
        self.files.extend(translations.files)

    return self


def _get_translations_multi_paths():
  """
  Returns the correct gettext translations that should be used for this
  request. This will never fail and return a dummy translation object
  if used outside of the request or if a translation cannot be found.
  """
  ctx = _request_ctx_stack.top
  if ctx is None:
      return None

  translations = getattr(ctx, 'babel_translations', None)
  if translations is None:
    babel_ext = ctx.app.extensions['babel']
    translations = None
    trs = None

    # reverse order: thus the application catalog is loaded last, so that
    # translations from libraries can be overriden
    for (dirname, domain) in reversed(babel_ext._translations_paths):
      trs = Translations.load(dirname,
                              locales=[flask_babel.get_locale()],
                              domain=domain)

      # babel.support.Translations is a subclass of
      # babel.support.NullTranslations, so we test if object has a 'merge'
      # method

      if not trs or not hasattr(trs, 'merge'):
        # got None or NullTranslations instance
        continue
      elif (translations is not None and hasattr(translations, 'merge')):
          translations.merge(trs)
      else:
          translations = trs

    # ensure translations is at least a NullTranslations object
    if translations is None:
      translations = trs

    ctx.babel_translations = translations

  return translations

# monkey patch flask-babel
flask_babel.get_translations = _get_translations_multi_paths

#: importable instance of :class:`Babel`
babel = Babel()


def localeselector():
  """
  Default locale selector used in abilian applications
  """
  # if a user is logged in, use the locale from the user settings
  user = getattr(g, 'user', None)
  if user is not None:
    locale = getattr(user, 'locale', None)
    if locale:
      return locale

  # Otherwise, try to guess the language from the user accept header the browser
  # transmits.  By default we support en/fr. The best match wins.
  return request.accept_languages.best_match(
    current_app.config['BABEL_ACCEPT_LANGUAGES']
  )


def timezoneselector():
  """
  Default timezone selector used in abilian applications
  """
  return LOCALTZ


@contextmanager
def set_locale(locale):
  """
  Change current locale.

  Can be used as a context manager to temporary change locale::

      with set_locale('fr') as fr_locale:
          ...

  :type locale: :class:`babel.core.Locale` or `str`
  :param locale: locale to use. If it's a string if must be a valid locale
                 specification
  :rtype: :class:`babel.core.Locale`
  :return: locale set
  """
  ctx = _request_ctx_stack.top
  if ctx is None:
    yield
    return

  if not isinstance(locale, Locale):
    locale = Locale.parse(locale)

  current_locale = getattr(ctx, 'babel_locale', None)
  ctx.babel_locale = locale
  yield locale
  ctx.babel_locale = current_locale


def get_template_i18n(template_name, locale):
  """
    Build template list with preceding locale if found
  """
  if locale is None:
    return [template_name]

  template_list = []
  parts = template_name.rsplit('.', 1)
  root = parts[0]
  suffix = parts[1]

  if locale.territory is not None:
    locale_string = '_'.join([locale.language, locale.territory])
    localized_template_path = u'.'.join([root, locale_string, suffix])
    template_list.append(localized_template_path)

  localized_template_path = u'.'.join([root, locale.language, suffix])
  template_list.append(localized_template_path)

  # append the default
  template_list.append(template_name)
  return template_list


class ensure_request_context(object):
  """
  Context manager that ensures a request context is set up
  """
  _rq_ctx = None

  def __enter__(self):
    if _request_ctx_stack.top is None:
      ctx = self._rq_ctx = current_app.test_request_context()
      ctx.__enter__()

  def __exit__(self, *args):
    ctx = self._rq_ctx
    self._rq_ctx = None

    if ctx is not None:
      ctx.__exit__(*args)


def render_template_i18n(template_name_or_list, **context):
  """
    Try to build an ordered list of template to satisfy the current locale
  """
  template_list = []
  # Use locale if present in **context
  if 'locale' in context:
    locale = Locale.parse(context['locale'])
  else:
    # Use get_locale() or default_locale
    locale = flask_babel.get_locale()

  if isinstance(template_name_or_list, string_types):
    template_list = get_template_i18n(template_name_or_list, locale)
  else:
    # Search for locale for each member of the list, do not bypass
    for template in template_name_or_list:
        template_list.extend(get_template_i18n(template, locale))

  with ensure_request_context(), set_locale(locale):
    return render_template(template_list, **context)


_NOT_WORD_RE = re.compile(r'[^\w\s]+', flags=re.UNICODE)


def to_lower_ascii(value):
  value = unicode(value)
  value = _NOT_WORD_RE.sub(u' ', value)
  value = unicodedata.normalize('NFKD', value)
  value = value.encode('ascii', 'ignore')
  value = value.decode('ascii')
  value = value.strip().lower()
  value = re.sub(r'[_\s]+', u' ', value)
  return value
